from ..Data._Rect import Rect
from ._CachedPlot import CachedPlot
from ..IO.Caching._CacheTarget import CacheTarget

class CachedPlotFactory(object):
    def __init__(self, cache_directory: str) -> None:
        self.__cache_directory: str = cache_directory

    def new(self, extent: Rect) -> CachedPlot:
        return CachedPlot(extent = extent)
    
    def load(self, target: CacheTarget) -> CachedPlot:
        return target.load_object(self.__cache_directory, CachedPlot)

    def save(self, plot: CachedPlot, target: CacheTarget) -> None:
        target.save_object(self.__cache_directory, plot)



from typing import cast

from QuasarCode._global_settings import settings_object as __settings_object
from QuasarCode.MPI import _raise_mpi_error
from .._MPI_Config_Base import _MPI_Config_Base

if not __settings_object.mpi_avalible:
    _raise_mpi_error()

from mpi4py import MPI



class _MPI_Config(_MPI_Config_Base[MPI.Intracomm]):
    """
    MPI Configuration

    (MPI enabled version)

    (readonly) comm

    (readonly) comm_size

    (readonly) rank

    (set; get) root

    (readonly) is_root

    check_is_root(int|None)

    update(comm-type, int|None)
    """

    def __init__(self, comm: MPI.Intracomm) -> None:
        super().__init__(
            comm = comm,
            size = comm.Get_size(),
            rank = comm.Get_rank(),
            root = 0
        )

if not _MPI_Config._is_singleton_avalible():
    _MPI_Config(MPI.COMM_WORLD)

mpi_config: _MPI_Config = cast(_MPI_Config, _MPI_Config._get_singleton())



from typing import TYPE_CHECKING, Any, Literal, Sequence, TypeVar, Generic
from abc import abstractmethod

from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from matplotlib.colorbar import Colorbar
from matplotlib.colorizer import ColorizingArtist
from matplotlib.colors import Colormap
from matplotlib.figure import Figure
if TYPE_CHECKING:
    from matplotlib.pylab import ArrayLike
from matplotlib.typing import ColorType, MarkerType
import numpy as np
from matplotlib.collections import PathCollection, PolyCollection
from matplotlib.container import ErrorbarContainer
from matplotlib.lines import Line2D
from matplotlib.contour import QuadContourSet
from matplotlib.image import AxesImage
from matplotlib.text import Text
from matplotlib.patches import Wedge

from ..Data._Rect import Rect
from ..Tools._Struct import CacheableStruct
from ..Tools._autoproperty import AutoProperty, AutoProperty_NonNullable
from ..Tools._CacheableFunction import CacheableFunction
from ._CachedPlotFontInfo import CachedPlotFontInfo

T = TypeVar("T")

class CachedPlotElement(CacheableStruct, Generic[T]):
    _result = AutoProperty[T](allow_uninitialised = True)
    def __init__(self, *cacheable_attributes: str, **kwargs):
        super().__init__(
            cacheable_attributes = cacheable_attributes,#list(cacheable_attributes) + ["_result"],
            **kwargs
        )
    @property
    def result(self) -> T|None:
        return self._result
    @abstractmethod
    def render(self, figure: Figure, axis: Axes, default_font: CachedPlotFontInfo, *args: Any, **kwargs: Any) -> None:
        """
        Render the element on the given figure and axis.
        
        Parameters:
            matplotlib.figure.Figure figure:
                The figure to render on.
            matplotlib.axes.Axes axis:
                The axis to render on.
        """
        raise NotImplementedError("Subclasses must implement this method.")

class CachedPlotLine(CachedPlotElement[Line2D]):
    x = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    y = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    label = AutoProperty[str](allow_uninitialised = True)
    colour = AutoProperty[ColorType](allow_uninitialised = True)
    linestyle = AutoProperty[str](allow_uninitialised = True)
    linewidth = AutoProperty[float](allow_uninitialised = True)
    alpha = AutoProperty_NonNullable[float](default_value = 1.0)
    def __init__(self, **kwargs):
        super().__init__("x", "y", "label", "colour", "linestyle", "linewidth", "alpha", **kwargs)
    def render(self, figure: Figure, axis: Axes, default_font: CachedPlotFontInfo, *args: Any, **kwargs: Any) -> None:
        self._result = axis.plot(
            self.x,
            self.y,
            label = self.label,
            color = self.colour,
            linestyle = self.linestyle,
            linewidth = self.linewidth,
            alpha = self.alpha,
            **kwargs
        )[0]
    @staticmethod
    def from_line(line: Line2D) -> "CachedPlotLine":
        coords: np.ndarray[tuple[int, int], np.dtype[np.floating]] = line.get_xydata()
        return CachedPlotLine(
            x = coords[:, 0],
            y = coords[:, 1],
            label = line.get_label(),
            colour = line.get_color(),
            linestyle = line.get_linestyle(),
            linewidth = line.get_linewidth(),
            alpha = line.get_alpha(),
            _result = line
        )

class CachedPlotScatter(CachedPlotElement[PathCollection]):
    x = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    y = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    label = AutoProperty[str](allow_uninitialised = True)
    colour = AutoProperty[ColorType|Sequence[ColorType]](allow_uninitialised = True)
    colourmap = AutoProperty[str|Colormap](allow_uninitialised = True)
    marker = AutoProperty[MarkerType](allow_uninitialised = True)
    size = AutoProperty["float|ArrayLike"](allow_uninitialised = True)
    alpha = AutoProperty["float|ArrayLike"](allow_uninitialised = True)
    def __init__(self, **kwargs):
        super().__init__("x", "y", "label", "colour", "colourmap", "marker", "size", "alpha", **kwargs)
    def render(self, figure: Figure, axis: Axes, default_font: CachedPlotFontInfo, *args: Any, **kwargs: Any):
        self._result = axis.scatter(self.x, self.y, c = self.colour, cmap = self.colourmap, marker = self.marker, s = self.size, alpha = self.alpha, label = self.label, **kwargs)
    @staticmethod
    def from_points(points: PathCollection) -> "CachedPlotScatter":
        coords: np.ndarray[tuple[int, int], np.dtype[np.floating]] = points.get_offsets()
        return CachedPlotScatter(
            x = coords[:, 0],
            y = coords[:, 1],
            label = points.get_label(),
            colour = points.get_facecolor(),
            colourmap = points.get_cmap(),
            marker = points.get_paths()[0],
            size = points.get_sizes(),
            alpha = points.get_alpha(),
            _result = points
        )

class CachedPlotErrorbar(CachedPlotElement[ErrorbarContainer]):
    x = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    y = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    xerr = AutoProperty[np.ndarray[tuple[int], np.dtype[np.floating]]](allow_uninitialised = True)
    yerr = AutoProperty[np.ndarray[tuple[int], np.dtype[np.floating]]](allow_uninitialised = True)
    label = AutoProperty[str](allow_uninitialised = True)
    line_colour = AutoProperty[ColorType|Sequence[ColorType]](allow_uninitialised = True)
    marker_colour = AutoProperty[ColorType|Sequence[ColorType]](allow_uninitialised = True)
    error_colour = AutoProperty[ColorType|Sequence[ColorType]](allow_uninitialised = True)
    error_line_width = AutoProperty[float](allow_uninitialised = True)
    colourmap = AutoProperty[str|Colormap](allow_uninitialised = True)
    marker = AutoProperty[MarkerType](allow_uninitialised = True, default_value = "")
    size = AutoProperty["float|ArrayLike"](allow_uninitialised = True)
    alpha = AutoProperty["float|ArrayLike"](allow_uninitialised = True)
    def __init__(self, **kwargs):
        super().__init__("x", "y", "xerr", "yerr", "colourmap", **kwargs)
    def render(self, figure: Figure, axis: Axes, default_font: CachedPlotFontInfo, *args: Any, **kwargs: Any):
        self._result = axis.errorbar(
            x = self.x,
            y = self.y,
            yerr = self.yerr,
            xerr = self.xerr,
            label = self.label,
            ecolor = self.error_colour,
            elinewidth = self.error_line_width,
            cmap = self.colourmap,
            marker = self.marker,
            markersize = self.size,
            alpha = self.alpha,
            color = self.line_colour,
            **kwargs
        )
    @staticmethod
    def from_errorbar(bars: ErrorbarContainer) -> "CachedPlotErrorbar":
        coords: np.ndarray[tuple[int, int], np.dtype[np.floating]] = bars.get_offsets()
        return CachedPlotErrorbar(
            x = coords[:, 0],
            y = coords[:, 1],
            label = bars.get_label(),
            colour = bars.get_facecolor(),
            colourmap = bars.get_cmap(),
            marker = bars.get_paths()[0],
            size = bars.get_sizes(),
            alpha = bars.get_alpha(),
            _result = bars
        )

class CachedPlotHexbin(CachedPlotElement[PolyCollection]):
    extent = AutoProperty_NonNullable[Rect]()
    gridsize = AutoProperty_NonNullable[int|tuple[int, int]]()
    polygon_offsets = AutoProperty_NonNullable[np.ndarray[tuple[int, int], np.dtype[np.floating]]]()
    bin_values = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    min_value = AutoProperty[float](allow_uninitialised = True)
    max_value = AutoProperty[float](allow_uninitialised = True)
    colourmap = AutoProperty[str|Colormap](allow_uninitialised = True)
    edgecolour = AutoProperty_NonNullable[ColorType|Literal["face", "none"]](default_value = "face")
    bin_alphas = AutoProperty[np.ndarray[tuple[int], np.dtype[np.floating]]](allow_uninitialised = True)
    def __init__(self, **kwargs):
        super().__init__("extent", "gridsize", "polygon_offsets", "bin_values", "min_value", "max_value", "colourmap", "edgecolour", "bin_alphas", **kwargs)
    def render(self, figure: Figure, axis: Axes, default_font: CachedPlotFontInfo, *args: Any, **kwargs: Any):
        self._result = axis.hexbin(
            x = self.polygon_offsets[:, 0],
            y = self.polygon_offsets[:, 1],
            C = self.bin_values,
            gridsize = self.gridsize,
            cmap = self.colourmap,
            vmin = self.min_value,
            vmax = self.max_value,
            edgecolor = self.edgecolour,
            extent = self.extent.extent,
            **kwargs
        )
        self._result.set_alpha(self.bin_alphas) # type: ignore[arg-type]
    @staticmethod
    def from_hexes(hexes: PolyCollection, extent: Rect, gridsize: int|tuple[int, int], min_value: float|None = None, max_value: float|None = None, edgecolours_match_faces: bool = False) -> "CachedPlotHexbin":
        n_hexes = hexes.get_array().shape[0]
        edgecolour = hexes.get_edgecolor()
        if edgecolours_match_faces:
            edgecolour = "face"
        elif isinstance(edgecolour, np.ndarray) and edgecolour.shape[0] == 0 and n_hexes > 0:
            edgecolour = "none"
        return CachedPlotHexbin(
            extent = extent,
            gridsize = gridsize,
            polygon_offsets = hexes.get_offsets(),
            bin_values = hexes.get_array(),
            min_value = min_value,
            max_value = max_value,
            edgecolour = edgecolour,
            colourmap = hexes.get_cmap(),
            bin_alphas = hexes.get_alpha(),
            _result = hexes
        )

class CachedPlotColourbar(CachedPlotElement[Colorbar]):
    target_element  = AutoProperty_NonNullable[str]()
    target_plot     = AutoProperty[str](allow_uninitialised = True)
    label           = AutoProperty[str](allow_uninitialised = True)
    add_to_axis     = AutoProperty_NonNullable[bool](default_value = True)
    location        = AutoProperty[Literal["left", "right", "top", "bottom"]|None](default_value = "right", allow_uninitialised = True)
    orientation     = AutoProperty[Literal["horizontal", "vertical", "top", "bottom"]](default_value = "vertical")
    extend          = AutoProperty[Literal["neither", "both", "min", "max"]](default_value = "neither")
    default_font    = AutoProperty_NonNullable[CachedPlotFontInfo]()
    label_font      = AutoProperty_NonNullable[CachedPlotFontInfo]()
    tick_label_font = AutoProperty_NonNullable[CachedPlotFontInfo]()
    def __init__(self, **kwargs):
        super().__init__("target_element", "target_plot", "label", "add_to_axis", "location", "orientation", "extend", "label_font", "tick_label_font", **kwargs)
        self.label_font = CachedPlotFontInfo()
    def render(self, figure: Figure, axis: Axes, target: ScalarMappable|ColorizingArtist, default_font: CachedPlotFontInfo, *args: Any, **kwargs: Any) -> None:
        """
        Render the element on the given figure and axis, using the target element.
        
        Parameters:
            matplotlib.figure.Figure figure:
                The figure to render on.
            matplotlib.axes.Axes axis:
                The axis to render on.
            ScalarMappable|ColorizingArtist target:
                The target element for the colourbar.
        """
        self._result = figure.colorbar(
            target,
            ax = axis if self.add_to_axis else None,
            cax = axis if not self.add_to_axis else None,
            location = self.location if self.add_to_axis else None,
            label = self.label,
            extend = self.extend,
            orientation = self.orientation if not self.add_to_axis else None,
            **kwargs
        )
        self._result.ax.xaxis.label.set_fontproperties(self.label_font.with_default(self.default_font).with_default(default_font).fontproperties)
        self._result.ax.yaxis.label.set_fontproperties(self.label_font.with_default(self.default_font).with_default(default_font).fontproperties)
        tick_label_font = self.tick_label_font.with_default(self.default_font).with_default(default_font)
        self._result.ax.tick_params(
            labelsize       = tick_label_font.size,
            labelcolor      = tick_label_font.color,
            labelfontfamily = tick_label_font.family,
        )

class CachedPlotContour(CachedPlotElement[QuadContourSet]):
    x                     = AutoProperty            [np.ndarray[tuple[int], np.dtype[np.floating]]]()
    y                     = AutoProperty            [np.ndarray[tuple[int], np.dtype[np.floating]]]()
    extent                = AutoProperty            [Rect                         ]()
    z                     = AutoProperty_NonNullable[np.ndarray[tuple[int, int], np.dtype[np.floating]]]()
    levels                = AutoProperty_NonNullable[tuple[float]                 ]()
    linewidths            = AutoProperty            ["ArrayLike"                  ](allow_uninitialised = True)
    linestyles            = AutoProperty            [tuple[Literal["solid", "dashed", "dashdot", "dotted"]]](allow_uninitialised = True)
    alpha_values          = AutoProperty            ["tuple[float|ArrayLike]"     ](allow_uninitialised = True)
    colours               = AutoProperty            [ColorType|tuple[ColorType]   ](allow_uninitialised = True)
    labeled_level_indexes = AutoProperty            [Sequence[int]                ](allow_uninitialised = True)
    label_positions       = AutoProperty            [Sequence[tuple[float, float]]](allow_uninitialised = True)
    label_format          = AutoProperty            [CacheableFunction            ](allow_uninitialised = True)
    font                  = AutoProperty_NonNullable[CachedPlotFontInfo]()
    def __init__(self, **kwargs):
        self.font = CachedPlotFontInfo()
        super().__init__("x", "y", "extent", "z", "levels", "linewidths", "linestyles", "alpha_values", "colours", "labeled_level_indexes", "label_positions", "label_format", "font", **kwargs)
    def render(self, figure: Figure, axis: Axes, default_font: CachedPlotFontInfo, *args: Any, **kwargs: Any) -> None:
        if sum([self.x is not None, self.y is not None]) == 1:
            raise ValueError("Both x and y must be provided together or not at all.")
        if self.x is None and self.extent is None:
            raise ValueError("If x and y are not provided, extent must be provided.")
        xy_args = [] if self.x is None else [self.x, self.y]
        self._result = axis.contour(
            *xy_args,
            self.z,
            levels = self.levels,
            linewidths = self.linewidths,
            linestyles = self.linestyles,
            alpha = self.alpha_values,
            colors = self.colours,
            **kwargs
        )
        if self.labeled_level_indexes is not None:
            labelled_levels = [self._result[-1].levels[i] for i in self.labeled_level_indexes]
            axis.clabel(
                CS     = self._result[-1],
                levels = labelled_levels,
                fmt    = [self.label_format(level) for level in labelled_levels] if self.label_format is not None else None,
                inline = 1,
                manual = self.label_positions,
                **self.font.with_default(default_font).fontdict,
            )
    @staticmethod
    def from_contours(contours: QuadContourSet, x: np.ndarray[tuple[int], np.dtype[np.floating]], y: np.ndarray[tuple[int], np.dtype[np.floating]], z: np.ndarray[tuple[int], np.dtype[np.floating]], uses_single_colour_value: bool = False) -> "CachedPlotContour":
        return CachedPlotContour(
            x = x,
            y = y,
            z = z,
            levels = tuple(contours.levels),
            linewidths = contours.get_linewidths(),
            linestyles = tuple(contours.linestyles),
            alpha_values = contours.alpha,
            colours = contours.colors if not uses_single_colour_value else tuple([contours.colors] * len(contours.levels)),
            _result = contours
        )

class CachedPlotImage(CachedPlotElement[AxesImage]):
    image = AutoProperty_NonNullable[np.ndarray[tuple[int, int], np.dtype[np.floating]]]()
    extent = AutoProperty_NonNullable[Rect]()
    origin = AutoProperty_NonNullable[Literal["upper", "lower"]](default_value = "upper")
    colourmap = AutoProperty[str|Colormap](allow_uninitialised = True)
    min_colour_value = AutoProperty[float](allow_uninitialised = True)
    max_colour_value = AutoProperty[float](allow_uninitialised = True)
    alpha = AutoProperty[np.ndarray[tuple[int, int], np.dtype[np.floating]]](allow_uninitialised = True)
    def __init__(self, **kwargs):
        super().__init__("image", "extent", "origin", "colourmap", "alpha", **kwargs)
    def render(self, figure: Figure, axis: Axes, default_font: CachedPlotFontInfo, *args: Any, **kwargs: Any):
        self._result = axis.imshow(
            self.image,
            extent = self.extent.extent,
            origin = self.origin,
            cmap = self.colourmap,
            vmin = self.min_colour_value,
            vmax = self.max_colour_value,
            alpha = self.alpha,
            **kwargs
        )
    @staticmethod
    def from_image(image: AxesImage) -> "CachedPlotImage":
        return CachedPlotImage(
            image = image.get_array(),
            extent = Rect.create_from_limits(*image.get_extent()),
            origin = image.origin,
            colourmap = image.get_cmap(),
            min_colour_value = image.get_clim()[0],
            max_colour_value = image.get_clim()[1],
            alpha = image.get_alpha()
        )

class CachedPlotText(CachedPlotElement[Text]):
    text          = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    x             = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    y             = AutoProperty_NonNullable[np.ndarray[tuple[int], np.dtype[np.floating]]]()
    font          = AutoProperty_NonNullable[CachedPlotFontInfo                           ]()
    colour        = AutoProperty            [ColorType                                    ]()
    alpha         = AutoProperty_NonNullable[float                                        ](default_value = 1.0)
    box_colour    = AutoProperty            [ColorType                                    ](allow_uninitialised = True)
    border_colour = AutoProperty            [ColorType                                    ](allow_uninitialised = True)
    def __init__(self, **kwargs):
        super().__init__("text", "x", "y", "font", "alpha", "box_colour", "border_colour", **kwargs)
        self.font = CachedPlotFontInfo()
    def render(self, figure: Figure, axis: Axes, default_font: CachedPlotFontInfo, *args: Any, **kwargs: Any) -> None:
        self._result = axis.text(
            x           = self.x,
            y           = self.y,
            s           = self.text,
            **self.font.with_default(default_font).fontdict,
            alpha       = self.alpha,
            bbox        = dict(facecolor = self.box_colour, edgecolor = self.border_colour) if (self.box_colour is not None or self.border_colour is not None) else None,
            **kwargs,
        )
    @staticmethod
    def from_text(text: Text) -> "CachedPlotText":
        coords: np.ndarray[tuple[int, int], np.dtype[np.floating]] = text.get_xydata()
        return CachedPlotText(
            x         = coords[:, 0],
            y         = coords[:, 1],
            label     = text.get_label(),
            colour    = text.get_color(),
            linestyle = text.get_linestyle(),
            linewidth = text.get_linewidth(),
            alpha     = text.get_alpha(),
        )

class CachedPlotPie(CachedPlotElement[tuple[list[Wedge], list[Text]] | tuple[list[Wedge], list[Text], list[Text]]]):
    values              = AutoProperty_NonNullable[Sequence[float]|Sequence[str]]()
    colours             = AutoProperty_NonNullable[Sequence[ColorType]|Sequence[str]]()
    texture             = AutoProperty_NonNullable[Sequence[Literal["/","\\","|","-","+", "x", "o", "O", ".", "*"]]]()
    shadow              = AutoProperty_NonNullable[bool](default_value = False)
    labels              = AutoProperty_NonNullable[Sequence[str]]()
    label_distance      = AutoProperty[float](default_value = 1.1, allow_uninitialised = True)
    rotate_labels       = AutoProperty_NonNullable[bool](default_value = False)
    percentage_format   = AutoProperty[str|CacheableFunction](allow_uninitialised = True)
    percentage_distance = AutoProperty_NonNullable[float](default_value = 0.6)
    centre              = AutoProperty_NonNullable[tuple[float,float]](default_value = (0, 0))
    radius              = AutoProperty_NonNullable[float](default_value = 1.0)
    start_angle         = AutoProperty_NonNullable[float](default_value = 0.0)
    segment_rotation    = AutoProperty_NonNullable[Literal["clockwise", "anticlockwise"]](default_value = "anticlockwise")
    wedge_properties    = AutoProperty[dict](allow_uninitialised = True)
    explode_distance    = AutoProperty[Sequence[float]|Sequence[str]](allow_uninitialised = True)
    font                = AutoProperty_NonNullable[CachedPlotFontInfo]()
    hide_axes           = AutoProperty_NonNullable[bool](default_value = True)
    ensure_complete_pie = AutoProperty_NonNullable[bool](default_value = True)
    labled_data         = AutoProperty[dict](allow_uninitialised = True)
    def __init__(self, **kwargs):
        super().__init__("values", "colours", "texture", "shadow", "labels", "label_distance", "rotate_labels", "percentage_format", "percentage_distance", "centre", "radius", "start_angle", "segment_rotation", "wedge_properties", "explode_distance", "font", "hide_axes", "ensure_complete_pie", "labled_data", **kwargs)
        self.font = CachedPlotFontInfo()
    def render(self, figure: Figure, axis: Axes, default_font: CachedPlotFontInfo, *args: Any, **kwargs: Any):
        if not self.ensure_complete_pie and sum(self.values) > 1:
            raise ValueError("The sum of the values must not exceed 1 when ensure_complete_pie is False.")
        self._result = axis.pie(
            x             = self.values,
            explode       = self.explode_distance,
            labels        = self.labels,
            colors        = self.colours,
            autopct       = self.percentage_format,
            pctdistance   = self.percentage_distance,
            shadow        = self.shadow,
            labeldistance = self.label_distance,
            startangle    = self.start_angle,
            radius        = self.radius,
            counterclock  = self.segment_rotation == "anticlockwise",
            wedgeprops    = self.wedge_properties,
            textprops     = self.font.with_default(default_font).fontdict,
            center        = self.centre,
            frame         = not self.hide_axes,
            rotatelabels  = self.rotate_labels,
            normalize     = self.ensure_complete_pie,
            hatch         = self.texture,
            data         = self.labled_data,
            **kwargs
        )
    @property
    def fractions(self) -> Sequence[float]:
        total = sum(self.values)
        return [value / total for value in self.values]
    @property
    def percentages(self) -> Sequence[float]:
        return [value * 100 for value in self.fractions]



import numpy as np
from matplotlib import pyplot as plt
from .fitline import LineOfBestFit, LineType

class DrawInstruction(object):
    """
    A data type for storing an instruction to draw a single item on a ManualFigure object.

    Arguments:
        deligate function -> A function with signiture (axis, *args, **kwargs) that edits the axis.
        list draw_args -> A list containing the function arguments.
        dict draw_kwargs -> A dictionary containing named function arguments.
    """

    def __init__(self, function, draw_args, draw_kwargs):
        self.__function = function
        self.__function_args = draw_args
        self.__function_kwargs = draw_kwargs

    def run(self, axis):
        """
        Executes the instruction an a provided axis.
        """
        self.__function(axis, *self.__function_args, **self.__function_kwargs)

    @staticmethod
    def plot(axis, *args, **kwargs):
        """
        Provided wrapper for the pyplot draw method.
        """
        axis.plot(*args, **kwargs)

    @staticmethod
    def errorbar(axis, *args, **kwargs):
        """
        Provided wrapper for the pyplot errorbar method.
        """
        axis.errorbar(*args, **kwargs)

    @staticmethod
    def scatter(axis, *args, **kwargs):
        """
        Provided wrapper for the pyplot scatter method.
        """
        axis.scatter(*args, **kwargs)

class IAutomaticFigure(object):
    """
    Interface for the Figure object.
    Any object that inherits this and defines "x", "y", "xError" and "yError" members
    that are publicly avalable will be treated as a valid object for use with the LineOfBestFit constructor.
    """

    def get_x(self):
        return self.x
    def get_y(self):
        return self.y
    def get_xError(self):
        return self.xError
    def get_yError(self):
        return self.yError

class ManualFigure(object):
    """
    A figure compatable with the QuasarCode.Science.Graph object that allows
    for a similar implementation of figures to the native matplotlib.pyplot one.
    Some added functionality has been made avalable.

    Arguments:
        tuple size -> A tuple with the horizontal and vertical size of the figure. (Defult is (6.4, 4.8))

        string title -> The title to be displayed on the graph. (Default is "")

        string xLabel -> The axis label to be displayed on the x-axis. (Default is "")

        string yLabel -> The axis label to be displayed on the y-axis. (Default is "")

        string saveName -> The png file name (can be precided with a file path) to save the graph under. (Default is "figure.png")
    """

    def __init__(self, size = None, title = "", xLabel = "", yLabel = "", saveName = "figure.png"):
        self.figure = plt.figure()
        self.size = size if size is not None else (6.4, 4.8)# This is the deafult as documented by https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.figure.html
        self.axis = None
        self.title = title
        self.xLabel = xLabel
        self.yLabel = yLabel
        self.saveName = saveName

        self.instructions = []

        self.plot()

    def plot(self, axisOveride = None):
        """
        Draws all the instructions on an axis.

        Arguments:
            matplotlib.axes._subplots.AxesSubplot axisOveride -> an alternitive axis to draw on.
        """
        if axisOveride is None:
            self.figure.clf()
            self.axis = self.figure.gca()
        else:
            self.axis = axisOveride

        for instruction in self.instructions:
            instruction.run(self.axis)

        self.axis.set_title(self.title)
        self.axis.set_xlabel(self.xLabel)
        self.axis.set_ylabel(self.yLabel)

        self.figure.set_figwidth(self.size[0])
        self.figure.set_figheight(self.size[1])

    def save(self, nameOveride = None):
        """
        Saves the figure in its current state as a png using the current save name.

        Arguments:
            string nameOveride -> The filename or filepath to use instead of the one stored by the object.
        """
        if nameOveride is not None and nameOveride != "":
            self.figure.savefig(nameOveride)
        elif self.saveName is not None and self.saveName != "":
            self.figure.savefig(self.saveName)

    def show(self):
        """
        Show the figure in a window.
        """
        self.figure.show()

    def addInstruction(self, function, *args, **kwargs):
        """
        Adds an instruction to the figure.
        This will be executed when the plot method is called.

        Arguments:
            deligate function -> the function to be executed that will edit the axis object.
                                See the static members of the DrawInstruction class for
                                some pre existing ones ready for use.

            Other arguments will be passed to the function specified.

            Other named arguments will be passed to the function specified.
        """
        self.instructions.append(DrawInstruction(function, args, kwargs))

    def clearInstructions(self):
        """
        Removes all the instructions.
        """
        self.instructions = []

    def addFit(self, x, y, xError = None, yError = None, lineType: LineType = LineType.straight, initialFitVariables: dict = None, customFitMethods: list = None, **kwargs):
        """
        Adds an instruction to draw a fit line using the data provided.

        Arguments:
            numpy.ndarray x -> The x coortinates of the data points.

            numpy.ndarray y -> The y coortinates of the data points.

            numpy.ndarray xError -> The errors on the x coordinates. Used to produce a horisontal errorbar.

            numpy.ndarray yError -> The errors on the y coordinates. Used to produce a vertical errorbar.
            
            LineType type -> The type of fit to plot. Options are from the LineType enum.

            dict initialFitVariables -> A dictionary containing the fit variables and their starting value used to
                                    overide he defults or to specify a custom fit. Variables should be in oposite
                                    order to that expected by the fitting functions i.e. m then c for y = m * x + c (Defult is None)

            list customFitMethods -> A list with two items: [the fitting function, the error function] (Defult is None)

            All other named arguments will be passed to the plot function used to draw the line.
        """
        if xError is not None:
            xError = np.array(xError)
        if yError is not None:
            yError = np.array(yError)
        fit = LineOfBestFit(ManualFigure.__FigureSpoofer(np.array(x), np.array(y), xError, yError), lineType, initialFitVariables, customFitMethods)

        if fit.fit_Y is not None:
            self.addInstruction(DrawInstruction.plot, [fit.fit_X, fit.fit_Y], *kwargs)

    class __FigureSpoofer(IAutomaticFigure):
        """
        A class that acts as a Figure object so the LineOfBestFit object can be created.
        """

        def __init__(self, x, y, dx, dy):
            self.x = x
            self.y = y
            self.xError = dx
            self.yError = dy

class Figure(IAutomaticFigure):
    """
    A figure object for encapsulating a matplotlib.pyplot graph working with a set of data and
    displaying it along with multiple optional lines of best fit.
    Automaticly formats the figure produced and allows overiding of all the usual pyplot features.

    Arguments:
        numpy.ndarray x -> The x coortinates of the data points.

        numpy.ndarray y -> The y coortinates of the data points.

        numpy.ndarray xError -> The errors on the x coordinates. Used to produce a horisontal errorbar.

        numpy.ndarray yError -> The errors on the y coordinates. Used to produce a vertical errorbar.

        tuple size -> A tuple with the horizontal and vertical size of the figure. (Defult is (6.4, 4.8))

        LineType fitLine -> The type of fit to plot. Options are from the LineType enum.
                            LineType.custom isn't valid - specify as None or as
                            LineType.none and then create the custom line seperately. (Defult is None)

        string title -> The title to be displayed on the graph. (Default is "")

        string xLabel -> The axis label to be displayed on the x-axis. (Default is "")

        string yLabel -> The axis label to be displayed on the y-axis. (Default is "")

        string saveName -> The png file name (can be precided with a file path) to save the graph under. (Default is "figure.png")
    """

    def __init__(self, x, y, xError = None, yError = None, size = None, fitLine = None, title = "", xLabel = "", yLabel = "", saveName = "figure.png"):
        self.figure = plt.figure(figsize = size)
        self.size = size if size is not None else [6.4, 4.8]# This is the deafult as documented by https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.figure.html
        self.axis = None
        self.title = title
        self.xLabel = xLabel
        self.yLabel = yLabel
        self.saveName = saveName
        
        self.__fitLineKwargs = []
        self.__dataPlotKwargs = []

        self.x: np.ndarray = np.array(x)
        self.y: np.ndarray = np.array(y)

        self.xError: np.ndarray = np.array(xError) if xError is not None else None
        self.yError: np.ndarray = np.array(yError) if yError is not None else None

        self.showErrorBars = True
        self.fitLines = []

        self.customXRange = False
        self.__xRange = (None, None)
        self.xRange = property(fget = (lambda self: self.__xRange), fset = self.setXRange)
        
        self.customYRange = False
        self.__yRange = (None, None)
        self.yRange = property(fget = lambda self: self.__yRange, fset = self.setYRange)

        if isinstance(fitLine, LineType) and fitLine != LineType.none and fitLine is not None:
            self.createFitLine(type = fitLine)

        self.plot()

    def plot(self, axisOveride = None, **kwargs):
        """
        Draws the figure onto an axis.

        Arguments:
            matplotlib.axes._subplots.AxesSubplot axisOveride -> an alternitive axis to draw on.

            All named arguments will be passed on to the plotting method of the axis (axis.plot or axis.errorbar).
        """
        self.__dataPlotKwargs = kwargs

        if axisOveride is None:
            self.figure.clf()
            self.axis = self.figure.gca()
        else:
            self.axis = axisOveride

        if self.showErrorBars and (self.xError is not None or self.yError is not None):
            # Add deafult styling to the kwargs if not allready present
            if "fmt" not in self.__dataPlotKwargs:
                self.__dataPlotKwargs["fmt"] = "r"
            if "linestyle" not in self.__dataPlotKwargs:
                self.__dataPlotKwargs["linestyle"] = ""

            # Plot the data with avalable errorbars
            self.axis.errorbar(self.x, self.y, self.yError, self.xError, **self.__dataPlotKwargs)

            # Ajust the x range
            if not self.customXRange:
                errorAjustX = [0, 0]
                if self.xError is not None:
                    errorAjustX[0] = self.xError[np.asarray(self.x == min(self.x)).nonzero()[0]]
                    errorAjustX[1] = self.xError[np.asarray(self.x == max(self.x)).nonzero()[0]]

                self.__xRange = (min(self.x) - errorAjustX[0], max(self.x) + errorAjustX[1])

            # Ajust the y range
            if not self.customYRange:
                errorAjustY = [0, 0]
                if self.yError is not None:
                    errorAjustY[0] = self.yError[np.asarray(self.y == min(self.y)).nonzero()[0]]
                    errorAjustY[1] = self.yError[np.asarray(self.y == max(self.y)).nonzero()[0]]

                self.__yRange = (min(self.y) - errorAjustY[0], max(self.y) + errorAjustY[1])

        else:
            # Plot the data
            self.axis.scatter(self.x, self.y, **self.__dataPlotKwargs)

            if not self.customXRange:
                self.__xRange = (min(self.x), max(self.x))
            if not self.customYRange:
                self.__yRange = (min(self.y), max(self.y))

        self.axis.set_xlim([self.__xRange[0], self.__xRange[1]])
        self.axis.set_ylim([self.__yRange[0], self.__yRange[1]])

        for i, line in enumerate(self.fitLines):
            if line.fit_Y is not None:
                self.axis.plot(line.fit_X, line.fit_Y, **self.__fitLineKwargs[i])

        self.axis.set_title(self.title)
        self.axis.set_xlabel(self.xLabel)
        self.axis.set_ylabel(self.yLabel)

        self.figure.set_figwidth(self.size[0])
        self.figure.set_figheight(self.size[1])

    def createFitLine(self, type: LineType = LineType.straight, clearOthers = True, initialFitVariables: dict = None, customFitMethods: list = None, **kwargs):
        """
        Creates a new line of best fit using the figure's data.

        Arguments:
            LineType type -> The type of fit to plot. Options are from the LineType enum.

            bool clearOthers -> Wether or not to clear all of the other lines of best fit. (Default is True)

            dict initialFitVariables -> A dictionary containing the fit variables and their starting value used to
                                    overide he defults or to specify a custom fit. Variables should be in oposite
                                    order to that expected by the fitting functions i.e. m then c for y = m * x + c (Defult is None)

            list customFitMethods -> A list with two items: [the fitting function, the error function] (Defult is None)

            All other named arguments will be passed to the plot function used to draw the line.
        """
        if clearOthers:
            self.fitLines = []
            self.__fitLineKwargs = []
        
        if type is not None:
            self.fitLines.append(LineOfBestFit(self, type, initialFitVariables, customFitMethods))
            self.__fitLineKwargs.append(kwargs)

    def save(self, nameOveride = None):
        """
        Saves the figure in its current state as a png using the current save name.

        Arguments:
            string nameOveride -> The filename or filepath to use instead of the one stored by the object.
        """
        if nameOveride is not None and nameOveride != "":
            self.figure.savefig(nameOveride)
        elif self.saveName is not None and self.saveName != "":
            self.figure.savefig(self.saveName)

    def show(self):
        """
        Show the figure in a window.
        """
        self.figure.show()


    # Property Accessors
    
    def setXRange(self, value: tuple):
        """
        Sets the range ov values on the x-axis
        """
        self.customXRange = True
        self.__xRange = value

    def setYRange(self, value: tuple):
        """
        Sets the range ov values on the y-axis
        """
        self.customYRange = True
        self.__yRange = value

    def getFittingOutput(self):
        """
        Retrives the output form all of the lines of best fit.
        """
        return "\n--------\n".join([fit.fittingOutput for fit in self.fitLines])


from ._CacheTarget import CacheTarget

class CacheTargetFactory(object):
    """
    A factory class for creating cache targets with common path templates.

    Parameters:
        str filename_template:
            The template for the (relative) file path of cache targets.
            It should contain format placeholders for the labels using the names of each label provided to 'label_names'.
            This path will be relative to the cache directory and, as such, should have no drive component or preceding "/".
        (args) tuple[str,...] label_names:
            The names of the labels that will be used to create the file path.

    Examples:
        ```
        ctf = CacheTargetFactory(
            "{software_name}/{run_name}/{name}.pickle",
            "software_name", "run_name", "name"
        )
        ctf.new(software_name="MySoftware", run_name="Run1", name="DataFile")
        # target a file at: {cache-directory}/MySoftware/Run1/DataFile.pickle
        ```
    """

    def __init__(self, filename_template: str, *label_names: str) -> None:
        self.__relative_filepath_template: str = filename_template
        self.__label_names: tuple[str, ...] = label_names

    def new(self, **labels: str) -> CacheTarget:
        """
        Creates a new CacheTarget instance using the provided labels.

        Parameters:
            (kwargs) dict[str,str] labels: A dictionary of label names and their corresponding values.
                Each label name specified in the constructor must be provided.

        Returns:
            CacheTarget -> A new CacheTarget instance.

        Raises:
            KeyError:
                If a required label is missing from the provided labels.
                See 'label_names' for the required labels.
        """
        for label_name in self.__label_names:
            if label_name not in labels:
                raise KeyError(f"No value provided for required label \"{label_name}\".")
        return CacheTarget(self.__relative_filepath_template.format(**{ label_name : labels[label_name] for label_name in self.__label_names }))

    @property
    def label_names(self) -> tuple[str, ...]:
        """
        The name of each label requiring an argument when creating a new cache target.

        Returns:
            tuple[str,...] -> A tuple containing the label names as strings.
        """
        return self.__label_names
    
    @property
    def template(self) -> str:
        """
        The template used to create the file path for the cache target.

        Returns:
            str -> The template string.
        """
        return self.__relative_filepath_template



from abc import ABC, abstractmethod
from functools import wraps
from typing import Generic, TypeVar, Union, Tuple, Iterable, Sized, Callable, Any, cast
from typing_extensions import ParamSpec, Concatenate

P = ParamSpec("P")
T = TypeVar("T")

class TypeShield_Base(Generic[T], ABC):

    @abstractmethod
    def report_format_types(self) -> str:
        raise NotImplementedError("Attempted to call abstract method.")

    @abstractmethod
    def check_type(self, value: Any) -> bool:
        raise NotImplementedError("Attempted to call abstract method.")

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError("Attempted to call abstract method.")

    def __str__(self) -> str:
        return self.report_format_types()
    
    def shield_check(self, value):
        """
        Check a value's type.
        """
        
        if self.check_type(value):
            return value
        raise TypeError(f"Attepted to assign value with incorrect type: {type(value).__qualname__} and any parent class(es) not in {self}.")

    def __call__(self, func: Callable[Concatenate[Any, P], T]):
        @wraps(func)
        def wrapper(value: Any, *args: P.args, **kwargs: P.kwargs) -> T:
            return func(self.shield_check(value), *args, **kwargs)
        return wrapper

class TypeShield(TypeShield_Base[T]):
    """
    Produces a function wrapper that restricts the type of the first argument.

    Many types may be specified.

    Usage:

    @TypeShield[Collection](tuple, list, numpy.ndarray, unyt.unyt_array)
    def my_func(val: Collection): pass
    """

    def __init__(self, *allowed_types: type):
        self.__allowed_types: Tuple[type, ...] = allowed_types

        # If nothing specified, apply a dummy guard type
        if len(self.__allowed_types) == 0:
            self.__allowed_types = (object,)

    @property
    def allowed_types(self) -> Tuple[type, ...]:
        """
        Types permitted by the TypeShield.
        """
        return self.__allowed_types
    
    def __len__(self) -> int:
         return 1

    def report_format_types(self) -> str:
        return f"TypeShield({', '.join([t.__qualname__ for t in self.__allowed_types])})"

    def check_type(self, value: Any) -> bool:
        for valid_target_type in self.__allowed_types:
            if isinstance(value, valid_target_type):
                return True
        return False

class NestedTypeShield(TypeShield_Base[T]):
    """
    Produces a function wrapper that restricts the type of the first argument by checking the types of the argument's substructure.
    For nested datatypes, this checks only the depth of the leftermost branches and NOT the whole dataset!
    For dictionary substructure, only test_dict[tuple(test_dict.keys())[0]] will be type checked.

    Many types may be specified at each level.

    Usage:

    @TypeShield[Collection[Union[Tuple[np.ndarray, ...], List[np.ndarray]]]](Collection, [tuple, list], np.ndarray, float)
    def my_func(val: Collection[Collection[np.ndarray]]): pass
    """

    def __init__(self, *allowed_type_layers: Union[type, Iterable[Union[type, None]], None]):
        self.__allowed_type_layers: Tuple[Tuple[Union[type, None], ...], ...] = tuple([(None,) if v is None else (v,) if isinstance(v, type) else tuple(v) if len(cast(Sized, v)) > 0 else (object,) for v in allowed_type_layers])

        # If nothing specified, apply a dummy guard type
        if len(self.__allowed_type_layers) == 0:
            self.__allowed_type_layers = ((object,),)

    @property
    def allowed_types(self) -> Tuple[Tuple[Union[type, None], ...], ...]:
        """
        Types permitted by the TypeShield.
        """
        return self.__allowed_type_layers
    
    def __len__(self) -> int:
         return len(self.__allowed_type_layers)

#    def report_format_types(self, subset = None) -> str:
#        if subset is None:
#            return "TypeShield(" + " -> ".join([(t.__qualname__ if t is not None else 'None') if (isinstance(t, type) or t is None) else (t[0].__qualname__ if t[0] is not None else 'None') if len(t) == 1 else self.report_format_types(subset = t) for t in self.__allowed_type_layers]) + ")"
#        else:
#            return "{" + ", ".join([(t.__qualname__ if t is not None else 'None') if (isinstance(t, type) or t is None) else (t[0].__qualname__ if t[0] is not None else 'None') if len(t) == 1 else self.report_format_types(subset = t) for t in subset]) + "}"
    def report_format_types(self, subset = None) -> str:
        return f"{'TypeShield(' if subset is None else '{'}{(' -> ' if subset is None else ', ').join([(t.__qualname__ if t is not None else 'None') if (isinstance(t, type) or t is None) else (t[0].__qualname__ if t[0] is not None else 'None') if len(t) == 1 else self.report_format_types(subset = t) for t in (self.__allowed_type_layers if subset is None else subset)])}{')' if subset is None else '}'}"

    def check_type(self, value: Any, level = 0) -> bool:
        valid = False
        for valid_target_type in self.__allowed_type_layers[level]:
            if isinstance(value, valid_target_type):
                valid = True
                break

        return (self.check_type(value[0] if not isinstance(value, dict) else value[tuple(value.keys())[0]], level = level + 1) if level < len(self) - 1 or len(value) == 0 else True) if valid else False



from abc import abstractmethod
from decimal import Decimal
import math
from typing import Iterable, TypeVar, Generic

import numpy as np

T = TypeVar("T")
class IRect(Generic[T]):
    """
    Interface for a 2D rectangle type.

    Use the following static classmethods on an implementation to create an instance:
        - create_from_size(x, y, width, height)
        - create_from_centre(x, y, width, height)
        - create_from_limits(x_min, x_max, y_min, y_max)
    """

    @classmethod
    @abstractmethod
    def create_from_size(cls, x: T, y: T, width: T, height: T) -> "IRect[T]":
        """
        Creates an IRect instance using the specified position of the bottom-left corner and size.

        Parameters:
            T x:
                The x-coordinate of the bottom-left corner of the rectangle.
            T y:
                The y-coordinate of the bottom-left corner of the rectangle.
            T width:
                The width of the rectangle.
            T height:
                The height of the rectangle.

        Returns:
            IRect[T] -> A new IRect instance defined by the given position and size.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @classmethod
    @abstractmethod
    def create_from_centre(cls, x: T, y: T, width: T, height: T) -> "IRect[T]":
        """
        Create a IRect instance using the centre coordinates and dimensions.

        Parameters:
            T x:
                The x-coordinate of the centre of the rectangle.
            T y:
                The y-coordinate of the centre of the rectangle.
            T width:
                The width of the rectangle.
            T height:
                The height of the rectangle.

        Returns:
            IRect[T] -> Instance centred on (x, y) with the specified width and height.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @classmethod
    @abstractmethod
    def create_from_limits(cls, x_min: T, x_max: T, y_min: T, y_max: T) -> "IRect[T]":
        """
        Create a IRect instance from the specified boundary limits.
        This is the same as the constructor parameters and is implemented for consistency.

        Parameters:
            T x_min:
                The minimum x-coordinate of the rectangle.
            T x_max:
                The maximum x-coordinate of the rectangle.
            T y_min:
                The minimum y-coordinate of the rectangle.
            T y_max:
                The maximum y-coordinate of the rectangle.

        Returns:
            IRect[T] -> A new instance of the Rect class defined by the given limits.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @classmethod
    @abstractmethod
    def create_from_data(cls, x_data: Iterable[T], y_data: Iterable[T]) -> "IRect[T]":
        """
        Create a IRect instance that encompasses the provided data.

        Parameters:
            Iterable[T] x_data:
                X-coordinates.
            Iterable[T] y_data:
                Y-coordinates.

        Returns:
            IRect[T] -> A new instance of the Rect class that encompasses the data.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def x(self) -> T:
        """
        The x-coordinate of the bottom-left corner of the rectangle.

        Returns:
            T -> The x-coordinate of the bottom-left corner.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @x.setter
    @abstractmethod
    def x(self, value: T) -> None:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def y(self) -> T:
        """
        The y-coordinate of the bottom-left corner of the rectangle.

        Returns:
            T -> The y-coordinate of the bottom-left corner.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @y.setter
    @abstractmethod
    def y(self, value: T) -> None:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def x1(self) -> T:
        """
        The x-coordinate of the top-right corner of the rectangle.

        Returns:
            T -> The x-coordinate of the top-right corner.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @x1.setter
    @abstractmethod
    def x1(self, value: T) -> None:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def y1(self) -> T:
        """
        The y-coordinate of the top-right corner of the rectangle.

        Returns:
            T -> The y-coordinate of the top-right corner.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @y1.setter
    @abstractmethod
    def y1(self, value: T) -> None:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def extent(self) -> tuple[T, T, T, T]:
        """
        The limits of the rectangle in the form (x_min, x_max, y_min, y_max).
        This corresponds to the format of a matplotlib 'extent' parameter.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def range(self) -> tuple[tuple[T, T], tuple[T, T]]:
        """
        The limits of the rectangle in the form ((x_min, x_max), (y_min, y_max)).
        This corresponds to the format of a numpy histogram2d 'range' parameter.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def width(self) -> T:
        """
        The width of the rectangle.

        Returns:
            T -> The width of the rectangle.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def height(self) -> T:
        """
        The height of the rectangle.

        Returns:
            T -> The height of the rectangle.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def centre(self) -> tuple[T, T]:
        """
        The coordinates of the center of the rectangle.

        Returns:
            tuple[T, T] -> (x, y)
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @property
    @abstractmethod
    def area(self) -> T:
        """
        The area of the rectangle.

        Returns:
            T -> The area of the rectangle.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __copy__(self) -> "IRect[T]":
        """
        Create a shallow copy of the rectangle.

        Returns:
            GenericRect[T] -> A new instance of the GenericRect class with the same properties.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __eq__(self, value) -> bool:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __ne__(self, value) -> bool:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __gt__(self, value) -> bool:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __ge__(self, value) -> bool:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __lt__(self, value) -> bool:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __le__(self, value) -> bool:
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __add__(self, value: "IRect[T]") -> "IRect[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __iadd__(self, value: "IRect[T]") -> "IRect[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __sub__(self, value: "IRect[T]") -> "IRect[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __isub__(self, value: "IRect[T]") -> "IRect[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __mul__(self, value: float) -> "IRect[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __imul__(self, value: float) -> "IRect[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __div__(self, value: float) -> "IRect[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __idiv__(self, value: float) -> "IRect[T]":
        raise NotImplementedError("This method must be implemented by subclasses.")



class GenericRect(IRect[T]):
    """
    A 2D rectangle that supports coordinates using integers or floating point values.

    Parameters:
        T x_min:
            X-coordinate of the bottom-left corner.
        T x_max:
            X-coordinate of the top-right corner.
        T y_min:
            Y-coordinate of the bottom-left corner.
        T y_max:
            Y-coordinate of the top-right corner.

    Use the following classmethods to create an instance:
        - create_from_size(x, y, width, height)
        - create_from_centre(x, y, width, height)
        - create_from_limits(x_min, x_max, y_min, y_max)
    """

    def __init__(self, x_min: T, x_max: T, y_min: T, y_max: T) -> None:
        self.__x_min = x_min
        self.__x_max = x_max
        self.__y_min = y_min
        self.__y_max = y_max

    @classmethod
    def create_from_size(cls, x: T, y: T, width: T, height: T) -> "GenericRect[T]":
        """
        Creates a GenericRect instance using the specified position of the bottom-left corner and size.

        Parameters:
            T x:
                The x-coordinate of the bottom-left corner of the rectangle.
            T y:
                The y-coordinate of the bottom-left corner of the rectangle.
            T width:
                The width of the rectangle.
            T height:
                The height of the rectangle.

        Returns:
            GenericRect[T] -> A new GenericRect instance defined by the given position and size.
        """
        return cls(x, x + width, y, y + height)

    @classmethod
    def create_from_centre(cls, x: T, y: T, width: T, height: T) -> "GenericRect[T]":
        """
        Create a GenericRect instance using the centre coordinates and dimensions.

        Parameters:
            T x:
                The x-coordinate of the centre of the rectangle.
            T y:
                The y-coordinate of the centre of the rectangle.
            T width:
                The width of the rectangle.
            T height:
                The height of the rectangle.

        Returns:
            GenericRect[T] -> Instance centred on (x, y) with the specified width and height.
        """
        return cls(x - width / 2, x + width / 2, y - height / 2, y + height / 2)

    @classmethod
    def create_from_limits(cls, x_min: T, x_max: T, y_min: T, y_max: T) -> "GenericRect[T]":
        """
        Create a GenericRect instance from the specified boundary limits.
        This is the same as the constructor parameters and is implemented for consistency.

        Parameters:
            T x_min:
                The minimum x-coordinate of the rectangle.
            T x_max:
                The maximum x-coordinate of the rectangle.
            T y_min:
                The minimum y-coordinate of the rectangle.
            T y_max:
                The maximum y-coordinate of the rectangle.

        Returns:
            GenericRect[T] -> A new instance of the Rect class defined by the given limits.
        """
        return cls(x_min, x_max, y_min, y_max)

    @classmethod
    def create_from_data(cls, x_data: Iterable[T], y_data: Iterable[T]) -> "GenericRect[T]":
        """
        Create a GenericRect instance that encompasses the provided data.

        Parameters:
            Iterable[T] x_data:
                X-coordinates.
            Iterable[T] y_data:
                Y-coordinates.

        Returns:
            GenericRect[T] -> A new instance of the Rect class that encompasses the data.
        """
        return cls(min(x_data), max(x_data), min(y_data), max(y_data))

    @property
    def x(self) -> T:
        """
        The x-coordinate of the bottom-left corner of the rectangle.

        Returns:
            T -> The x-coordinate of the bottom-left corner.
        """
        return self.__x_min

    @x.setter
    def x(self, value: T) -> None:
        self.__x_min = value

    @property
    def y(self) -> T:
        """
        The y-coordinate of the bottom-left corner of the rectangle.

        Returns:
            T -> The y-coordinate of the bottom-left corner.
        """
        return self.__y_min

    @y.setter
    def y(self, value: T) -> None:
        self.__y_min = value

    @property
    def x1(self) -> T:
        """
        The x-coordinate of the top-right corner of the rectangle.

        Returns:
            T -> The x-coordinate of the top-right corner.
        """
        return self.__x_max

    @x1.setter
    def x1(self, value: T) -> None:
        self.__x_max = value

    @property
    def y1(self) -> T:
        """
        The y-coordinate of the top-right corner of the rectangle.

        Returns:
            T -> The y-coordinate of the top-right corner.
        """
        return self.__y_max

    @y1.setter
    def y1(self, value: T) -> None:
        self.__y_max = value

    @property
    def extent(self) -> tuple[T, T, T, T]:
        """
        The limits of the rectangle in the form (x_min, x_max, y_min, y_max).
        This corresponds to the format of a matplotlib 'extent' parameter.
        """
        return (self.__x_min, self.__x_max, self.__y_min, self.__y_max)

    @property
    def range(self) -> tuple[tuple[T, T], tuple[T, T]]:
        """
        The limits of the rectangle in the form ((x_min, x_max), (y_min, y_max)).
        This corresponds to the format of a numpy histogram2d 'range' parameter.
        """
        return ((self.__x_min, self.__x_max), (self.__y_min, self.__y_max))

    @property
    def width(self) -> T:
        """
        The width of the rectangle.

        Returns:
            T -> The width of the rectangle.
        """
        return self.__x_max - self.__x_min

    @property
    def height(self) -> T:
        """
        The height of the rectangle.
        
        Returns:
            T -> The height of the rectangle.
        """
        return self.__y_max - self.__y_min

    @property
    def centre(self) -> tuple[T, T]:
        """
        The coordinates of the center of the rectangle.

        Returns:
            tuple[T, T] -> (x, y)
        """
        return ((self.__x_max + self.__x_min) / 2, (self.__y_max + self.__y_min) / 2)

    @property
    def area(self) -> T:
        """
        The area of the rectangle.

        Returns:
            T -> The area of the rectangle.
        """
        return self.width * self.height

    def __copy__(self) -> "GenericRect[T]":
        """
        Create a shallow copy of the rectangle.

        Returns:
            GenericRect[T] -> A new instance of the GenericRect class with the same properties.
        """
        return GenericRect(self.__x_min, self.__x_max, self.__y_min, self.__y_max)

    def __str__(self) -> str:
        return f"Rect(x: {self.__x_min} -> {self.__x_max}, y: {self.__y_min} -> {self.__y_max})"

    def __repr__(self) -> str:
        return "GenericRect: " + self.__str__()

    def __eq__(self, value):
        if isinstance(value, GenericRect):
            return self.__x_min == value.__x_min and self.__x_max == value.__x_max and self.__y_min == value.__y_min and self.__y_max == value.__y_max
        return False

    def __ne__(self, value) -> bool:
        return not self.__eq__(value)

    def __gt__(self, value) -> bool:
        if isinstance(value, IRect):
            return self.area > value.area
        elif isinstance(value, (int, float, Decimal, np.integer, np.floating)):
            return self.area > value
        return False

    def __ge__(self, value) -> bool:
        if isinstance(value, IRect):
            return self.area >= value.area
        elif isinstance(value, (int, float, Decimal, np.integer, np.floating)):
            return self.area >= value
        return False

    def __lt__(self, value) -> bool:
        if isinstance(value, IRect):
            return self.area < value.area
        elif isinstance(value, (int, float, Decimal, np.integer, np.floating)):
            return self.area < value
        return False

    def __le__(self, value) -> bool:
        if isinstance(value, IRect):
            return self.area <= value.area
        elif isinstance(value, (int, float, Decimal, np.integer, np.floating)):
            return self.area <= value
        return False
    
    def __add__(self, value: IRect[T]) -> "GenericRect[T]":
        return type(self).create_from_size(
            self.__x_min + value.x,
            self.__y_min + value.y,
            self.width + value.width,
            self.height + value.height
        )

    def __iadd__(self, value: IRect[T]) -> "GenericRect[T]":
        current_width = self.width
        current_height = self.height
        self.__x_min += value.x
        self.__y_min += value.y
        self.__x_max = self.__x_min + current_width + value.width
        self.__y_max = self.__y_min + current_height + value.height
        return self

    def __sub__(self, value: IRect[T]) -> "GenericRect[T]":
        return type(self).create_from_size(
            self.__x_min - value.x,
            self.__y_min - value.y,
            self.width - value.width,
            self.height - value.height
        )

    def __isub__(self, value: IRect[T]) -> "GenericRect[T]":
        current_width = self.width
        current_height = self.height
        self.__x_min -= value.x
        self.__y_min -= value.y
        self.__x_max = self.__x_min + current_width - value.width
        self.__y_max = self.__y_min + current_height - value.height
        return self

    def __mul__(self, value: float) -> "GenericRect[T]":
        return type(self).create_from_size(
            self.__x_min,
            self.__y_min,
            self.width * value,
            self.height * value
        )

    def __imul__(self, value: float) -> "GenericRect[T]":
        self.__x_max = self.__x_min + (self.width * value)
        self.__y_max = self.__y_min + (self.height * value)
        return self

    def __div__(self, value: float) -> "GenericRect[T]":
        return type(self).create_from_size(
            self.__x_min,
            self.__y_min,
            self.width / value,
            self.height / value
        )

    def __idiv__(self, value: float) -> "GenericRect[T]":
        self.__x_max = self.__x_min + (self.width / value)
        self.__y_max = self.__y_min + (self.height / value)
        return self



class Rect(GenericRect[float]):
    """
    A 2D rectangle.

    Parameters:
        float x_min:
            X-coordinate of the bottom-left corner.
        float x_max:
            X-coordinate of the top-right corner.
        float y_min:
            Y-coordinate of the bottom-left corner.
        float y_max:
            Y-coordinate of the top-right corner.

    Use the following classmethods to create an instance:
        - create_from_size(x, y, width, height)
        - create_from_centre(x, y, width, height)
        - create_from_limits(x_min, x_max, y_min, y_max)
    """

    @classmethod
    def create_from_size(cls, x: float, y: float, width: float, height: float) -> "Rect":
        """
        Creates a Rect instance using the specified position of the bottom-left corner and size.

        Parameters:
            float x:
                The x-coordinate of the bottom-left corner of the rectangle.
            float y:
                The y-coordinate of the bottom-left corner of the rectangle.
            float width:
                The width of the rectangle.
            float height:
                The height of the rectangle.

        Returns:
            Rect -> A new Rect instance defined by the given position and size.
        """
        return super().create_from_size(x, width, y, height)

    @classmethod
    def create_from_centre(cls, x: float, y: float, width: float, height: float) -> "Rect":
        """
        Create a Rect instance using the centre coordinates and dimensions.

        Parameters:
            float x:
                The x-coordinate of the centre of the rectangle.
            float y:
                The y-coordinate of the centre of the rectangle.
            float width:
                The width of the rectangle.
            float height:
                The height of the rectangle.

        Returns:
            Rect -> Instance centred on (x, y) with the specified width and height.
        """
        return super().create_from_centre(x, y, width, height)

    @classmethod
    def create_from_limits(cls, x_min: float, x_max: float, y_min: float, y_max: float) -> "Rect":
        """
        Create a Rect instance from the specified boundary limits.
        This is the same as the constructor parameters and is implemented for consistency.

        Parameters:
            float x_min:
                The minimum x-coordinate of the rectangle.
            float x_max:
                The maximum x-coordinate of the rectangle.
            float y_min:
                The minimum y-coordinate of the rectangle.
            float y_max:
                The maximum y-coordinate of the rectangle.

        Returns:
            Rect -> A new instance of the Rect class defined by the given limits.
        """
        return super().create_from_limits(x_min, x_max, y_min, y_max)

    @classmethod
    def create_from_data(cls, x_data: Iterable[float], y_data: Iterable[float]) -> "Rect":
        """
        Create a Rect instance that encompasses the provided data.

        Parameters:
            Iterable[float] x_data:
                X-coordinates.
            Iterable[float] y_data:
                Y-coordinates.

        Returns:
            Rect -> A new instance of the Rect class that encompasses the data.
        """
        return super().create_from_data(x_data, y_data)

    @property
    def x(self) -> float:
        """
        The x-coordinate of the bottom-left corner of the rectangle.

        Returns:
            float -> The x-coordinate of the bottom-left corner.
        """
        return GenericRect.x.fget(self)

    @x.setter
    def x(self, value: float) -> None:
        return GenericRect.x.fset(self, value)

    @property
    def y(self) -> float:
        """
        The y-coordinate of the bottom-left corner of the rectangle.

        Returns:
            float -> The y-coordinate of the bottom-left corner.
        """
        return GenericRect.y.fget(self)

    @y.setter
    def y(self, value: float) -> None:
        return GenericRect.y.fset(self, value)

    @property
    def x1(self) -> float:
        """
        The x-coordinate of the top-right corner of the rectangle.

        Returns:
            float -> The x-coordinate of the top-right corner.
        """
        return GenericRect.x1.fget(self)

    @x1.setter
    def x1(self, value: float) -> None:
        return GenericRect.x1.fset(self, value)

    @property
    def y1(self) -> float:
        """
        The y-coordinate of the top-right corner of the rectangle.

        Returns:
            float -> The y-coordinate of the top-right corner.
        """
        return GenericRect.y1.fget(self)

    @y1.setter
    def y1(self, value: float) -> None:
        return GenericRect.y1.fset(self, value)

    @property
    def extent(self) -> tuple[float, float, float, float]:
        """
        The limits of the rectangle in the form (x_min, x_max, y_min, y_max).
        This corresponds to the format of a matplotlib 'extent' parameter.
        """
        return GenericRect.extent.fget(self)

    @property
    def range(self) -> tuple[tuple[float, float], tuple[float, float]]:
        """
        The limits of the rectangle in the form ((x_min, x_max), (y_min, y_max)).
        This corresponds to the format of a numpy histogram2d 'range' parameter.
        """
        return GenericRect.range.fget(self)

    @property
    def width(self) -> float:
        """
        The width of the rectangle.

        Returns:
            float -> The width of the rectangle.
        """
        return GenericRect.width.fget(self)

    @property
    def height(self) -> float:
        """
        The height of the rectangle.
        
        Returns:
            float -> The height of the rectangle.
        """
        return GenericRect.height.fget(self)

    @property
    def centre(self) -> tuple[float, float]:
        """
        The coordinates of the center of the rectangle.

        Returns:
            tuple[float, float] -> (x, y)
        """
        return GenericRect.centre.fget(self)

    @property
    def area(self) -> float:
        """
        The area of the rectangle.

        Returns:
            float -> The area of the rectangle.
        """
        return  GenericRect.area.fget(self)

    def __copy__(self) -> "Rect":
        """
        Create a shallow copy of the rectangle.

        Returns:
            Rect -> A new instance of the Rect class with the same properties.
        """
        return Rect(*self.extent)

    def __repr__(self) -> str:
        return "Rect: " + self.__str__()

    def __add__(self, value: IRect[T]) -> "Rect":
        return super().__add__(value)

    def __sub__(self, value: IRect[T]) -> "Rect":
        return super().__sub__(value)

    def __mul__(self, value: float) -> "Rect":
        return super().__mul__(value)

    def __div__(self, value: float) -> "Rect":
        return super().__div__(value)



class DiscreteRect(GenericRect[int]):
    """
    A 2D rectangle.

    Parameters:
        int x_min:
            X-coordinate of the bottom-left corner.
        int x_max:
            X-coordinate of the top-right corner.
        int y_min:
            Y-coordinate of the bottom-left corner.
        int y_max:
            Y-coordinate of the top-right corner.

    Use the following classmethods to create an instance:
        - create_from_size(x, y, width, height)
        - create_from_centre(x, y, width, height)
        - create_from_limits(x_min, x_max, y_min, y_max)
    """

    @classmethod
    def create_from_size(cls, x: int, y: int, width: int, height: int) -> "DiscreteRect":
        """
        Creates a DiscreteRect instance using the specified position of the bottom-left corner and size.

        Parameters:
            int x:
                The x-coordinate of the bottom-left corner of the rectangle.
            int y:
                The y-coordinate of the bottom-left corner of the rectangle.
            int width:
                The width of the rectangle.
            int height:
                The height of the rectangle.

        Returns:
            DiscreteRect -> A new DiscreteRect instance defined by the given position and size.
        """
        return super().create_from_size(x, y, width, height)
    
    @classmethod
    def create_from_centre(cls, x: int, y: int, width: int, height: int) -> "DiscreteRect":
        """
        Create a DiscreteRect instance using the centre coordinates and dimensions.
        In the event that the width or height are odd values, the upper bounds will be further from the centre by 1 then the lower bounds.

        Parameters:
            int x:
                The x-coordinate of the centre of the rectangle.
            int y:
                The y-coordinate of the centre of the rectangle.
            int width:
                The width of the rectangle.
            int height:
                The height of the rectangle.

        Returns:
            DiscreteRect -> Instance centred on (x, y) with the specified width and height.
        """
        return cls(math.floor(x - width / 2), math.ceil(x + width / 2), math.floor(y - height / 2), math.ceil(y + height / 2))
    
    @classmethod
    def create_from_limits(cls, x_min: int, x_max: int, y_min: int, y_max: int) -> "DiscreteRect":
        """
        Create a DiscreteRect instance from the specified boundary limits.
        This is the same as the constructor parameters and is implemented for consistency.

        Parameters:
            int x_min:
                The minimum x-coordinate of the rectangle.
            int x_max:
                The maximum x-coordinate of the rectangle.
            int y_min:
                The minimum y-coordinate of the rectangle.
            int y_max:
                The maximum y-coordinate of the rectangle.

        Returns:
            DiscreteRect -> A new instance of the DiscreteRect class defined by the given limits.
        """
        return super().create_from_limits(x_min, x_max, y_min, y_max)
    
    @classmethod
    def create_from_data(cls, x_data: Iterable[int], y_data: Iterable[int]) -> "DiscreteRect":
        """
        Create a DiscreteRect instance that encompasses the provided data.

        Parameters:
            Iterable[int] x_data:
                X-coordinates.
            Iterable[int] y_data:
                Y-coordinates.

        Returns:
            DiscreteRect -> A new instance of the DiscreteRect class that encompasses the data.
        """
        return super().create_from_data(x_data, y_data)

    @property
    def x(self) -> int:
        """
        The x-coordinate of the bottom-left corner of the rectangle.

        Returns:
            int -> The x-coordinate of the bottom-left corner.
        """
        return GenericRect.x.fget(self)

    @x.setter
    def x(self, value: int) -> None:
        return GenericRect.x.fset(self, value)

    @property
    def y(self) -> int:
        """
        The y-coordinate of the bottom-left corner of the rectangle.

        Returns:
            int -> The y-coordinate of the bottom-left corner.
        """
        return GenericRect.y.fget(self)

    @y.setter
    def y(self, value: int) -> None:
        return GenericRect.y.fset(self, value)

    @property
    def x1(self) -> int:
        """
        The x-coordinate of the top-right corner of the rectangle.

        Returns:
            int -> The x-coordinate of the top-right corner.
        """
        return GenericRect.x1.fget(self)

    @x1.setter
    def x1(self, value: int) -> None:
        return GenericRect.x1.fset(self, value)

    @property
    def y1(self) -> int:
        """
        The y-coordinate of the top-right corner of the rectangle.

        Returns:
            int -> The y-coordinate of the top-right corner.
        """
        return GenericRect.y1.fget(self)

    @y1.setter
    def y1(self, value: int) -> None:
        return GenericRect.y1.fset(self, value)

    @property
    def extent(self) -> tuple[int, int, int, int]:
        """
        The limits of the rectangle in the form (x_min, x_max, y_min, y_max).
        This corresponds to the format of a matplotlib 'extent' parameter.
        """
        return GenericRect.extent.fget(self)

    @property
    def range(self) -> tuple[tuple[int, int], tuple[int, int]]:
        """
        The limits of the rectangle in the form ((x_min, x_max), (y_min, y_max)).
        This corresponds to the format of a numpy histogram2d 'range' parameter.
        """
        return GenericRect.range.fget(self)

    @property
    def width(self) -> int:
        """
        The width of the rectangle.

        Returns:
            int -> The width of the rectangle.
        """
        return GenericRect.width.fget(self)

    @property
    def height(self) -> int:
        """
        The height of the rectangle.
        
        Returns:
            int -> The height of the rectangle.
        """
        return GenericRect.height.fget(self)

    @property
    def centre(self) -> tuple[int, int]:
        """
        The coordinates of the center of the rectangle.

        Returns:
            tuple[int, int] -> (x, y)
        """
        return (math.floor((self.x1 + self.x) / 2), math.floor((self.y1 + self.y) / 2))

    @property
    def area(self) -> int:
        """
        The area of the rectangle.

        Returns:
            int -> The area of the rectangle.
        """
        return  GenericRect.area.fget(self)

    def __copy__(self) -> "DiscreteRect":
        """
        Create a shallow copy of the rectangle.

        Returns:
            DiscreteRect -> A new instance of the DiscreteRect class with the same properties.
        """
        return DiscreteRect(*self.extent)

    def __repr__(self) -> str:
        return "DiscreteRect: " + self.__str__()

    def __add__(self, value: IRect[T]) -> "DiscreteRect":
        return super().__add__(value)

    def __sub__(self, value: IRect[T]) -> "DiscreteRect":
        return super().__sub__(value)

    def __mul__(self, value: float) -> "DiscreteRect":
        return type(self).create_from_size(
            self.x,
            self.y,
            math.floor(self.width * value),
            math.floor(self.height * value)
        )

    def __imul__(self, value: float) -> "DiscreteRect":
        self.x1 = self.x + math.floor(self.width * value)
        self.y1 = self.y + math.floor(self.height * value)
        return self

    def __div__(self, value: float) -> "DiscreteRect":
        return type(self).create_from_size(
            self.x,
            self.y,
            math.floor(self.width / value),
            math.floor(self.height / value)
        )

    def __idiv__(self, value: float) -> "DiscreteRect":
        self.x1 = self.x + math.floor(self.width / value)
        self.y1 = self.y + math.floor(self.height / value)
        return self



from abc import ABC, abstractmethod
from typing import Any, TypeVar, Type

T = TypeVar("T", bound = "Cacheable")

class Cacheable(ABC):
    """
    Marks an object as being cacheable using a CacheTarget.

    Abstract Methods:
        __from_cache_data__
        __get_cache_data__
    """

    @classmethod
    @abstractmethod
    def __from_cache_data__(cls: Type[T], data: dict[str, Any]) -> T:
        """
        Load the object from a cache target.

        Parameters:
            dict[str, Any] data:
                The data to load the object from.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")

    @abstractmethod
    def __get_cache_data__(self) -> dict[str, Any]:
        """
        Get the data to be cached.

        Returns:
            dict[str, Any] -> The data to be cached.
        """
        raise NotImplementedError("This method must be implemented by subclasses.")



from enum import Enum

from ..Data._Rect import DiscreteRect

class ScreenResolution(Enum):
    """
    16:9 Screen Resolutions:
        nHD     :    640 x   360 = (2/3)K
        FWVGA   :    854 x   480 = (427/480)K
        qHD     :    960 x   540 = 1K
        HD      :  1,280 x   720 = (4/3)K
        WXGA    :  1,366 x   768 = (683/480)K
        HD+     :  1,600 x   900 = (5/3)K
        Full HD :  1,920 x 1,080 = 2K
        QHD     :  2,560 x 1,440 = (8/3)K
        QHD+    :  3,200 x 1,800 = (10/3)K
        4K UHD  :  3,840 x 2,160 = 4K
        5K      :  5,120 x 2,880 = 5K
        8K UHD  :  7,680 x 4,320 = 8K
        16K UHD : 15,360 x 8,640 = 16K
    """

    nHD       = DiscreteRect.create_from_size(0, 0, 640, 360)
    FWVGA     = DiscreteRect.create_from_size(0, 0, 854, 480)
    qHD       = DiscreteRect.create_from_size(0, 0, 960, 540)
    HD        = DiscreteRect.create_from_size(0, 0, 1280, 720)
    WXGA      = DiscreteRect.create_from_size(0, 0, 1366, 768)
    HD_PLUS   = DiscreteRect.create_from_size(0, 0, 1600, 900)
    FULL_HD   = DiscreteRect.create_from_size(0, 0, 1920, 1080)
    QHD       = DiscreteRect.create_from_size(0, 0, 2560, 1440)
    QHD_PLUS  = DiscreteRect.create_from_size(0, 0, 3200, 1800)
    FOUR_K    = DiscreteRect.create_from_size(0, 0, 3840, 2160)
    FIVE_K    = DiscreteRect.create_from_size(0, 0, 5120, 2880)
    EIGHT_K   = DiscreteRect.create_from_size(0, 0, 7680, 4320)
    SIXTEEN_K = DiscreteRect.create_from_size(0, 0, 15360, 8640)

    @staticmethod
    def get_all():
        """
        Returns all screen resolutions.
        """
        return ScreenResolution.__members__.values()

    @staticmethod
    def get_16_by_9():
        """
        Returns all 16:9 screen resolutions.
        """
        return [res for name, res in ScreenResolution.__members__ if name in (
            "nHD",
            "FWVGA",
            "qHD",
            "HD",
            "WXGA",
            "HD_PLUS",
            "FULL_HD",
            "QHD",
            "QHD_PLUS",
            "FOUR_K",
            "FIVE_K",
            "EIGHT_K",
            "SIXTEEN_K"
        )]

    @staticmethod
    def get_720p():
        """
        Returns the HD resolution (1280x720).
        """
        return ScreenResolution.HD

    @staticmethod
    def get_1080p():
        """
        Returns the Full HD resolution (1920x1080).
        """
        return ScreenResolution.FULL_HD

    @staticmethod
    def get_4K():
        """
        Returns the 4K resolution (3840x2160).
        """
        return ScreenResolution.FOUR_K



import numpy as np
from decimal import Decimal

#def consistancyTest(a, b, aErr, bErr = 0):
#    """
#    Performs a consistancy test to determine the integer number of sigma within which \
#two values lie relitive to each other and their errors.
#    """
#    sigma = 0
#    consistant = False
#    while not consistant:
#        sigma += 1
#        checkValue = np.abs(b - a) - sigma * np.sqrt(bErr**2 + aErr**2)
#        #print(checkValue)
#        consistant = checkValue <= 0
#    return sigma

def consistancy_test(a, b, aErr, bErr = 0):#TODO: fully test this aproach
    """
    Performs a consistancy test to determine the integer number of sigma within which \
two values lie relitive to each other and their errors.
    """
    return int(np.ceil(np.abs(b - a) / np.sqrt(bErr**2 + aErr**2)))

def is_consistant(a, b, aErr, bErr, sigma):
    """
    Convenience method for checking if two values are consistant within a given sigma value
    """
    return consistancy_test(a, b, aErr, bErr) <= sigma

def error_string(value: float, error: float, units: str = None, normalise: bool = True):#TODO: sort normalisation for value
    """
    Formats a number and its error as a single string
    """
    if normalise:#TODO: this dosn't normalise the value it just does the error
        error = sig_fig(error, 1)

    if units is None:
        return "{} \u00B1 {}".format(value, float(error))
    else:
        return "({} \u00B1 {}) {}".format(value, error, units)

def sig_fig(num: Decimal, digits: int):
    """
    Rounds a number to a set number of significant figures
    """
    num = Decimal(num)

    if np.abs(num) >= 1:
        power = 0
        while np.abs(num) > 10:
            power += 1
            num /= 10

    else:
        power = 0
        while np.abs(num) < 1:
            power -= 1
            num *= 10

    # Return number to 10^digits * 10^0
    num = true_round(num * 10**(Decimal(digits - 1)), 0) * 10**-(Decimal(digits - 1)) * 10**Decimal(power)

    return num

def standard_form(num: Decimal):
    """
    Determines the value and power for the standard form representation of a number
    """
    num = Decimal(num)
    power = 0

    if num >= 1:
        new_num = None
        new_power = None
        while num > 1:
            new_num = num
            new_power = power

            new_power += 1
            new_num /= 10

            if new_num < 1:
                break
            else:
                num = new_num
                power = new_power

    else:
        while num < 1:
            power -= 1
            num *= 10
            if num > 1:
                break

    return num, power

def standard_form_string(num: Decimal, numberSigFig = None, latex = False):
    """
    Determines the standard form representation of a number and represents it as a string
    """
    number, power = standard_form(num)

    template = "${}\\times10^{{{}}}$" if latex else "{} * 10^{}"

    return template.format(float(sig_fig(number, numberSigFig) if numberSigFig is not None else number), power) if power != 0 else str(float(sig_fig(number, numberSigFig) if numberSigFig is not None else number))

def standard_form_string_if_nessessary(num: Decimal, numberSigFig = None):
    if np.abs(num) >= 9999 or np.abs(num) < 0.001:
        return standard_form_string(num, numberSigFig)
    else:
        return str(float(sig_fig(num, numberSigFig) if numberSigFig is not None else num))

def true_round(number: Decimal, nDigits: int = None):
    """
    Rounds a number according to common rounding principles.
    i.e. flaw if next digit less than 5 or celing if more than or equal to 5
    """
    number = Decimal(number)

    if nDigits is None:
        nDigits = 0

    normal = number * 10**nDigits

    rest = normal - int(normal)

    if rest >= 0.5:
        return (normal - rest + 1) * 10**-nDigits
    else:
        return (normal - rest) * 10**-nDigits


import numpy as np
from typing import Union, Collection, List

class _ArrayVisuliserElement(object):

    def __init__(self, title: Union[str, None], data: np.ndarray):
        self.__title = title if title is not None else ""
        self.__data = np.array(data)
        self.__dimensions = len(self.__data.shape)
        self.__count = np.prod(self.__data.shape)
        self.__minimal_count_string = f"{self.__count}" + ("" if self.__dimensions == 1 else f" ({', '.join([str(v) for v in self.__data.shape])})")

        self.__required_title_characters = len(self.__title)
        self.__min_required_count_characters = int(np.log10(self.__count)) + 1 + (0 if self.__dimensions == 1 else (3 + sum(int(np.log10(l)) + 1 for l in self.__data.shape) + (2 * (self.__dimensions - 1))))
        self.__max_count_characters_total_count = int(np.log10(self.__count)) + 1
        self.__max_count_characters_per_dimension = [int(np.log10(dimension_size)) + 1 for dimension_size in self.__data.shape]

        self.__percent = lambda n: 100 * n / self.__count

        self.__n_nan                      = np.isnan(  self.__data                                       )
        self.__n_negitive_inf             =         (  self.__data == -np.inf                            )
        self.__n_positive_inf             =         (  self.__data ==  np.inf                            )
        self.__n_negitive                 =         ( (self.__data  <  0)     & (self.__data != -np.inf) )
        self.__n_zero                     =         (  self.__data ==  0                                 )
        self.__n_positive                 =         ( (self.__data  >  0)     & (self.__data !=  np.inf) )
        self.__n_negitive_1_to_below_zero =         ( (self.__data >= -1)     & (self.__data <  0)       )
        self.__n_above_zero_to_positive_1 =         ( (self.__data <=  1)     & (self.__data >  0)       )
        self.__n_negitive_e_to_below_zero =         ( (self.__data >= -np.e)  & (self.__data <  0)       )
        self.__n_above_zero_to_positive_e =         ( (self.__data <=  np.e)  & (self.__data >  0)       )

        if self.__dimensions == 1:
            self.__n_nan                      = self.__n_nan.sum()
            self.__n_negitive_inf             = self.__n_negitive_inf.sum()
            self.__n_positive_inf             = self.__n_positive_inf.sum()
            self.__n_negitive                 = self.__n_negitive.sum()
            self.__n_zero                     = self.__n_zero.sum()
            self.__n_positive                 = self.__n_positive.sum()
            self.__n_negitive_1_to_below_zero = self.__n_negitive_1_to_below_zero.sum()
            self.__n_above_zero_to_positive_1 = self.__n_above_zero_to_positive_1.sum()
            self.__n_negitive_e_to_below_zero = self.__n_negitive_e_to_below_zero.sum()
            self.__n_above_zero_to_positive_e = self.__n_above_zero_to_positive_e.sum()
        else:
            axes = np.array(np.arange(self.__dimensions), dtype = int)
            summation_axis_tuples = [
                tuple(axes[axes != axis_index])
                for axis_index
                in axes
            ]

            self.__n_nan                      = (self.__n_nan.sum(),                      [(self.__n_nan.sum(                      axis = summation_axis_tuples[i]) > 0).sum() for i in range(len(summation_axis_tuples))])
            self.__n_negitive_inf             = (self.__n_negitive_inf.sum(),             [(self.__n_negitive_inf.sum(             axis = summation_axis_tuples[i]) > 0).sum() for i in range(len(summation_axis_tuples))])
            self.__n_positive_inf             = (self.__n_positive_inf.sum(),             [(self.__n_positive_inf.sum(             axis = summation_axis_tuples[i]) > 0).sum() for i in range(len(summation_axis_tuples))])
            self.__n_negitive                 = (self.__n_negitive.sum(),                 [(self.__n_negitive.sum(                 axis = summation_axis_tuples[i]) > 0).sum() for i in range(len(summation_axis_tuples))])
            self.__n_zero                     = (self.__n_zero.sum(),                     [(self.__n_zero.sum(                     axis = summation_axis_tuples[i]) > 0).sum() for i in range(len(summation_axis_tuples))])
            self.__n_positive                 = (self.__n_positive.sum(),                 [(self.__n_positive.sum(                 axis = summation_axis_tuples[i]) > 0).sum() for i in range(len(summation_axis_tuples))])
            self.__n_negitive_1_to_below_zero = (self.__n_negitive_1_to_below_zero.sum(), [(self.__n_negitive_1_to_below_zero.sum( axis = summation_axis_tuples[i]) > 0).sum() for i in range(len(summation_axis_tuples))])
            self.__n_above_zero_to_positive_1 = (self.__n_above_zero_to_positive_1.sum(), [(self.__n_above_zero_to_positive_1.sum( axis = summation_axis_tuples[i]) > 0).sum() for i in range(len(summation_axis_tuples))])
            self.__n_negitive_e_to_below_zero = (self.__n_negitive_e_to_below_zero.sum(), [(self.__n_negitive_e_to_below_zero.sum( axis = summation_axis_tuples[i]) > 0).sum() for i in range(len(summation_axis_tuples))])
            self.__n_above_zero_to_positive_e = (self.__n_above_zero_to_positive_e.sum(), [(self.__n_above_zero_to_positive_e.sum( axis = summation_axis_tuples[i]) > 0).sum() for i in range(len(summation_axis_tuples))])
        
        self.__minimum_non_infinate_value =                                                                             self.__data[(self.__data != -np.inf) & (~np.isnan(self.__data))                     ].min() if ((self.__data != -np.inf) & (~np.isnan(self.__data))).sum() > 0 else None
        self.__maximum_non_infinate_value =                                                                             self.__data[(self.__data !=  np.inf) & (~np.isnan(self.__data))                     ].max() if ((self.__data !=  np.inf) & (~np.isnan(self.__data))).sum() > 0 else None
        self.__smallest_nonzero_value     = self.__data[(self.__data != 0) & (~np.isnan(self.__data))].flatten()[np.abs(self.__data[(self.__data !=  0)      & (~np.isnan(self.__data))].flatten()).argmin()]       if ((self.__data !=  0     ) & (~np.isnan(self.__data))).sum() > 0 else None

    @property
    def _required_title_characters(self):
        return self.__required_title_characters
    @property
    def _min_required_count_characters(self):
        return self.__min_required_count_characters

    def __str__(self):
        return f"{self.__title} | {self.__minimal_count_string}"

    def render_title(self, line_length: int, internal: bool) -> str:
        sep = "-" if internal else ","

        if self.__title.strip() != "":
            n_extra_dashes = line_length - 6 - len(self.__title)
            n_dashes_left =  int(n_extra_dashes / 2)
            n_dashes_right =  (int(n_extra_dashes / 2) + (0 if n_extra_dashes % 2 == 0 else 1))
            return f"{sep * (n_dashes_left + 2)} {self.__title} {sep * (n_dashes_right + 2)}"

        else:
            return sep * line_length

    def render_body(self, max_count_char_allowence: int, line_length: int, show_plus_minus_1: bool, show_plus_minus_e: bool, show_min_max: bool, max_unused_decimal_places_for_minmax: int) -> str:
        row_padding = " " * _ArrayVisuliserElement._calculate_row_padding(max_count_char_allowence, line_length)

        multi_dimensional_count_format_insert = None
        if self.__dimensions > 1:
            multi_dimensional_count_format_insert = (" " * (max_count_char_allowence - self.__min_required_count_characters)) + f"{{:{self.__max_count_characters_total_count}.0f}}" + " (" + ", ".join([f"{{:{self.__max_count_characters_per_dimension[i]}.0f}}" for i in range(self.__dimensions)]) + ")"

        lines = []

        # Number
        lines.append(
            f"    N    :         : {{:{max_count_char_allowence}.0f}} {row_padding}".format(self.__count)
            if self.__dimensions == 1 else
            f"    N    :         : {' ' * (max_count_char_allowence - self.__min_required_count_characters)}{self.__minimal_count_string} {row_padding}"
        )

        # +inf
        lines.append(
            f"    \u221E    : {{:6.2f}}% : {{:{max_count_char_allowence}.0f}} {row_padding}".format(self.__percent(self.__n_positive_inf), self.__n_positive_inf)
            if self.__dimensions == 1 else
            f"    \u221E    : {{:6.2f}}% : {multi_dimensional_count_format_insert} {row_padding}".format(self.__percent(self.__n_positive_inf[0]), self.__n_positive_inf[0], *self.__n_positive_inf[1])
        )

        # Positive
        lines.append(
            f"   +ve   : {{:6.2f}}% : {{:{max_count_char_allowence}.0f}} {row_padding}".format(self.__percent(self.__n_positive), self.__n_positive)
            if self.__dimensions == 1 else
            f"   +ve   : {{:6.2f}}% : {multi_dimensional_count_format_insert} {row_padding}".format(self.__percent(self.__n_positive[0]), self.__n_positive[0], *self.__n_positive[1])
        )

        # Zero
        lines.append(
            f"    0    : {{:6.2f}}% : {{:{max_count_char_allowence}.0f}} {row_padding}".format(self.__percent(self.__n_zero), self.__n_zero)
            if self.__dimensions == 1 else
            f"    0    : {{:6.2f}}% : {multi_dimensional_count_format_insert} {row_padding}".format(self.__percent(self.__n_zero[0]), self.__n_zero[0], *self.__n_zero[1])
        )

        # Negitive
        lines.append(
            f"   -ve   : {{:6.2f}}% : {{:{max_count_char_allowence}.0f}} {row_padding}".format(self.__percent(self.__n_negitive), self.__n_negitive)
            if self.__dimensions == 1 else
            f"   -ve   : {{:6.2f}}% : {multi_dimensional_count_format_insert} {row_padding}".format(self.__percent(self.__n_negitive[0]), self.__n_negitive[0], *self.__n_negitive[1])
        )

        # -inf
        lines.append(
            f"   -\u221E    : {{:6.2f}}% : {{:{max_count_char_allowence}.0f}} {row_padding}".format(self.__percent(self.__n_negitive_inf), self.__n_negitive_inf)
            if self.__dimensions == 1 else
            f"   -\u221E    : {{:6.2f}}% : {multi_dimensional_count_format_insert} {row_padding}".format(self.__percent(self.__n_negitive_inf[0]), self.__n_negitive_inf[0], *self.__n_negitive_inf[1])
        )

        # NaN
        lines.append(
            f"   nan   : {{:6.2f}}% : {{:{max_count_char_allowence}.0f}} {row_padding}".format(self.__percent(self.__n_nan), self.__n_nan)
            if self.__dimensions == 1 else
            f"   nan   : {{:6.2f}}% : {multi_dimensional_count_format_insert} {row_padding}".format(self.__percent(self.__n_nan[0]), self.__n_nan[0], *self.__n_nan[1])
        )

        if show_plus_minus_1:
        
            # Blank line to seperate sections
            lines.append(
                f"         :         : {' ' * max_count_char_allowence} {row_padding}"
            )

            # 1 >= v > 0
            lines.append(
                f" +1\u2265x> 0 : {{:6.2f}}% : {{:{max_count_char_allowence}.0f}} {row_padding}".format(self.__percent(self.__n_above_zero_to_positive_1), self.__n_above_zero_to_positive_1)
                if self.__dimensions == 1 else
                f" +1\u2265x> 0 : {{:6.2f}}% : {multi_dimensional_count_format_insert} {row_padding}".format(self.__percent(self.__n_above_zero_to_positive_1[0]), self.__n_above_zero_to_positive_1[0], *self.__n_above_zero_to_positive_1[1])
            )

            # 0 > v >= -1
            lines.append(
                f"  0>x\u2265-1 : {{:6.2f}}% : {{:{max_count_char_allowence}.0f}} {row_padding}".format(self.__percent(self.__n_negitive_1_to_below_zero), self.__n_negitive_1_to_below_zero)
                if self.__dimensions == 1 else
                f"  0>x\u2265-1 : {{:6.2f}}% : {multi_dimensional_count_format_insert} {row_padding}".format(self.__percent(self.__n_negitive_1_to_below_zero[0]), self.__n_negitive_1_to_below_zero[0], *self.__n_negitive_1_to_below_zero[1])
            )

        if show_plus_minus_e:
        
            # Blank line to seperate sections
            lines.append(
                f"         :         : {' ' * max_count_char_allowence} {row_padding}"
            )

            # 1 >= v > 0
            lines.append(
                f" +e\u2265x> 0 : {{:6.2f}}% : {{:{max_count_char_allowence}.0f}} {row_padding}".format(self.__percent(self.__n_above_zero_to_positive_e), self.__n_above_zero_to_positive_e)
                if self.__dimensions == 1 else
                f" +e\u2265x> 0 : {{:6.2f}}% : {multi_dimensional_count_format_insert} {row_padding}".format(self.__percent(self.__n_above_zero_to_positive_e[0]), self.__n_above_zero_to_positive_e[0], *self.__n_above_zero_to_positive_e[1])
            )

            # 0 > v >= -1
            lines.append(
                f"  0>x\u2265-e : {{:6.2f}}% : {{:{max_count_char_allowence}.0f}} {row_padding}".format(self.__percent(self.__n_negitive_e_to_below_zero), self.__n_negitive_e_to_below_zero)
                if self.__dimensions == 1 else
                f"  0>x\u2265-e : {{:6.2f}}% : {multi_dimensional_count_format_insert} {row_padding}".format(self.__percent(self.__n_negitive_e_to_below_zero[0]), self.__n_negitive_e_to_below_zero[0], *self.__n_negitive_e_to_below_zero[1])
            )

        if show_min_max:
            
            if self.__maximum_non_infinate_value is None and self.__smallest_nonzero_value is None and self.__minimum_non_infinate_value is None:
                fmt = "{}"
            else:
                max_chars = max_count_char_allowence + 10
                n_chars_for_int_portion = max(1, 1 + int( np.log10(max(np.abs(self.__maximum_non_infinate_value) if (self.__maximum_non_infinate_value is not None and self.__maximum_non_infinate_value != 0) else 1, np.abs(self.__smallest_nonzero_value) if (self.__smallest_nonzero_value is not None and self.__smallest_nonzero_value != 0) else 1, np.abs(self.__minimum_non_infinate_value) if (self.__minimum_non_infinate_value is not None and self.__minimum_non_infinate_value != 0) else 1))))
                min_dp_for_precision    =        1 + int(-np.log10(min(np.abs(self.__maximum_non_infinate_value) if (self.__maximum_non_infinate_value is not None and self.__maximum_non_infinate_value != 0) else 1, np.abs(self.__smallest_nonzero_value) if (self.__smallest_nonzero_value is not None and self.__smallest_nonzero_value != 0) else 1, np.abs(self.__minimum_non_infinate_value) if (self.__minimum_non_infinate_value is not None and self.__minimum_non_infinate_value != 0) else 1)))
                n_dp = max(min(max_unused_decimal_places_for_minmax, max_chars - n_chars_for_int_portion - 1), min_dp_for_precision)
                fmt = f"{{:{max_chars}.{n_dp}f}}"
        
            # Blank line to seperate sections
            lines.append(
                f"         :         : {' ' * max_count_char_allowence} {row_padding}"
            )

            # 1 >= v > 0
            insert_value = fmt.format(self.__maximum_non_infinate_value)
            lines.append(
                f"max   \u2260 \u221E: {' ' * (max_count_char_allowence - len(insert_value) + 10)}{insert_value} {row_padding}"
                if self.__maximum_non_infinate_value is not None else
                f"max   \u2260 \u221E: N/A       {' ' * max_count_char_allowence} {row_padding}"
            )

            # 1 >= v > 0
            insert_value = fmt.format(self.__smallest_nonzero_value)
            lines.append(
                f"small \u2260 0: {' ' * (max_count_char_allowence - len(insert_value) + 10)}{insert_value} {row_padding}"
                if self.__smallest_nonzero_value is not None else
                f"small \u2260 0: N/A       {' ' * max_count_char_allowence} {row_padding}"
            )

            # 0 > v >= -1
            insert_value = fmt.format(self.__minimum_non_infinate_value)
            lines.append(
                f"min   \u2260-\u221E: {' ' * (max_count_char_allowence - len(insert_value) + 10)}{insert_value} {row_padding}"
                if self.__minimum_non_infinate_value is not None else
                f"min   \u2260-\u221E: N/A       {' ' * max_count_char_allowence} {row_padding}"
            )

        # Blank line at bottom
        lines.append(
            f"         :         : {' ' * max_count_char_allowence} {row_padding}"
        )

        return "\n".join(lines)

    @staticmethod
    def render_bottom(line_length: int) -> str:
        return "-" * line_length

    @staticmethod
    def render_empty_title(line_length: int) -> str:
        return " " * line_length

    @staticmethod
    def render_empty_body(line_length: int, show_plus_minus_1: bool, show_plus_minus_e: bool, show_min_max: bool) -> str:
        return "\n".join([" " * line_length] * (8 + (3 if show_plus_minus_1 else 0) + (3 if show_plus_minus_e else 0) + (4 if show_min_max else 0)))
    
    @staticmethod
    def get_smallest_body_line_length():
        return 22 + 1

    @staticmethod
    def calculate_min_line_length(max_title_char_allowence: int, max_count_char_allowence: int):
        return max(6 + max_title_char_allowence, 22 + max_count_char_allowence)

    @staticmethod
    def _calculate_row_padding(max_count_char_allowence: int, line_length: int):
        return max(0, line_length - (22 + max_count_char_allowence))

class ArrayVisuliser(object):

    def __init__(self,
                 datasets: Union[np.ndarray, Collection[Union[np.ndarray, None]], Collection[Collection[Union[np.ndarray, None]]]],
                 titles: Union[str, Collection[Union[str, None]], Collection[Collection[Union[str, None]]], None] = None,
                 number_missing_titles = True):

        # Is there just one dataset?
        if not isinstance(datasets, list):
            datasets = [[datasets]]
            if titles is not None:
                titles = [[titles]]

        # Is there just one row of datasets?
        elif not isinstance(datasets[0], list):
            datasets = [datasets]
            if titles is not None:
                titles = [titles]

        self.__n_width = len(datasets[0])
        self.__n_height = len(datasets)
        self.__elements: List[List[Union[_ArrayVisuliserElement, None]]] = [
            [
                (_ArrayVisuliserElement(titles[row][col] if titles is not None else (f"(#{row * self.__n_width + col})" if number_missing_titles else None),
                                       datasets[row][col])
                if datasets[row][col] is not None else
                None)
                for col
                in range(self.__n_width)
            ]
            for row
            in range(self.__n_height)
        ]

        self.__number_of_elements = self.__n_width * self.__n_height

        self.__count = sum([
                                sum([
                                        1 if self.__elements[i][j] is not None else 0
                                        for j
                                        in range(self.__n_width)
                                ])
                                for i
                                in range(self.__n_height)
                       ])

    def __len__(self):
        return self.__count
    
    @property
    def number_of_cells(self):
        return self.__number_of_elements

    def __str__(self):
        return "\n".join(["\n".join([element.__str__() for element in row]) for row in self.__elements])

    def render(self, all_columns_same_width = False, min_count_allowence: int = 0, show_plus_minus_1: bool = False, show_plus_minus_e: bool = False, show_min_max: bool = True, max_unused_decimal_places_for_minmax: int = 3):
        result = None

        column_count_chars = [
            max(max([
                    (self.__elements[row][col]._min_required_count_characters
                    if self.__elements[row][col] is not None else
                    1)
                    for row
                    in range(self.__n_height)
                ]),
                min_count_allowence)
            for col
            in range(self.__n_width)
        ]

        column_line_lengths = [
            max([
                (_ArrayVisuliserElement.calculate_min_line_length(self.__elements[row][col]._required_title_characters, column_count_chars[col])
                if self.__elements[row][col] is not None else
                _ArrayVisuliserElement.get_smallest_body_line_length())
                for row
                in range(self.__n_height)
            ])
            for col
            in range(self.__n_width)
        ]
        if all_columns_same_width:
            column_line_lengths = [max(column_line_lengths)] * len(column_line_lengths)

        row_strings = []
        for row in range(self.__n_height):
            is_top_row = row == 0
            is_bottom_row = row == self.__n_height - 1

            rendered_element_lines = []
            for col in range(self.__n_width):
                rendered_element_lines.append((self.__elements[row][col].render_body(column_count_chars[col], column_line_lengths[col], show_plus_minus_1, show_plus_minus_e, show_min_max, max_unused_decimal_places_for_minmax) if self.__elements[row][col] is not None else _ArrayVisuliserElement.render_empty_body(column_line_lengths[col], show_plus_minus_1, show_plus_minus_e, show_min_max)).split("\n"))

            total_lines_per_render = len(rendered_element_lines[0])

            row_lines = [""] * (total_lines_per_render + (1 if not is_bottom_row else 2))
            for col in range(self.__n_width):
                is_populated_dataset = self.__elements[row][col] is not None
#                is_dataset_to_left = col > 0 and self.__elements[row][col - 1] is not None
                is_dataset_to_right = col < self.__n_width - 1 and self.__elements[row][col + 1] is not None
                is_dataset_above = not is_top_row and self.__elements[row - 1][col] is not None
#                is_dataset_below = not is_bottom_row and self.__elements[row + 1][col] is not None
                is_dataset_diagonal_above_and_right = not is_top_row and col < self.__n_width - 1 and self.__elements[row - 1][col + 1] is not None

                if col == 0:
                    row_lines[0] += "#" if is_populated_dataset or is_dataset_above else " "
                row_lines[0] += self.__elements[row][col].render_title(column_line_lengths[col], not is_dataset_above) if is_populated_dataset else _ArrayVisuliserElement.render_bottom(column_line_lengths[col]) if is_dataset_above else _ArrayVisuliserElement.render_empty_title(column_line_lengths[col])
                row_lines[0] += "#" if is_populated_dataset or is_dataset_to_right or is_dataset_above or is_dataset_diagonal_above_and_right else " "

                for i in range(total_lines_per_render):
                    if col == 0:
                        row_lines[i + 1] += "|" if is_populated_dataset else " "
                    row_lines[i + 1] += rendered_element_lines[col][i]
                    row_lines[i + 1] += ("¦" if is_populated_dataset else "|") if is_dataset_to_right else ("|" if is_populated_dataset else " ")

                if is_bottom_row:
                    if col == 0:
                        row_lines[-1] += "#" if is_populated_dataset else " "
                    row_lines[-1] += _ArrayVisuliserElement.render_bottom(column_line_lengths[col]) if is_populated_dataset else _ArrayVisuliserElement.render_empty_title(column_line_lengths[col])
                    row_lines[-1] += "#" if is_populated_dataset or is_dataset_to_right else " "

            row_strings.append("\n".join(row_lines))

        result = "\n".join(row_strings)

        return result

    def print(self, *args, all_columns_same_width = False, min_count_allowence: int = 0, show_plus_minus_1: bool = False, show_plus_minus_e: bool = False, show_min_max: bool = True, **kwargs):
        print(self.render(all_columns_same_width = all_columns_same_width,
                          min_count_allowence = min_count_allowence,
                          show_plus_minus_1 = show_plus_minus_1,
                          show_plus_minus_e = show_plus_minus_e,
                          show_min_max = show_min_max),
              *args,
              **kwargs)

    @staticmethod
    def arrange(columns: int, datasets: Collection[np.ndarray], titles: Union[List[Union[str, None]], None] = None):
        if not isinstance(columns, int):
            raise TypeError(f"Number of columns must be an integer not {type(columns)}")
        if columns < 1:
            raise ValueError(f"Number of columns must be nonzero and positive but was {columns}")
        if not isinstance(datasets, Collection):
            raise TypeError("More than one dataset must be passed.")
        if titles is not None and len(datasets) != len(titles):
            raise ValueError("Mismatched number of datasets and titles.")

        n_datasets = len(datasets)
        n_rows = int(n_datasets / columns) + (0 if n_datasets % columns == 0 else 1)

        return ArrayVisuliser([[(datasets[i * columns + j] if (i * columns + j < n_datasets) else None) for j in range(columns)] for i in range(n_rows)],
                              [[(  titles[i * columns + j] if (i * columns + j < n_datasets) else None) for j in range(columns)] for i in range(n_rows)] if titles is not None else None)

    @staticmethod
    def row(*datasets: np.ndarray, titles: Union[List[Union[str, None]], None] = None):
        if len(datasets) == 0:
            raise TypeError("More than one dataset must be passed.")
        if titles is not None and len(datasets) != len(titles):
            raise ValueError("Mismatched number of datasets and titles.")
        return ArrayVisuliser(datasets = [datasets], titles = [titles] if titles is not None else None)

    @staticmethod
    def column(*datasets: np.ndarray, titles: Union[List[Union[str, None]], None] = None):
        if len(datasets) == 0:
            raise TypeError("More than one dataset must be passed.")
        if titles is not None and len(datasets) != len(titles):
            raise ValueError("Mismatched number of datasets and titles.")
        return ArrayVisuliser(datasets = [[d] for d in datasets],
                              titles = [[t] for t in titles] if titles is not None else None)



class IDice(object):
    """
    A dice that can be rolled to produce an integer result starting from 1
    """
    @property
    def Sides(self) -> int:
        """
        Number of sides on the dice
        """
        raise NotImplementedError()

    def Roll(self):
        """
        Rolls the dice
        """
        raise NotImplementedError()


from typing import Any

import numpy as np

from ..IO.Caching import Cacheable

class Bins(Cacheable):
    def __init__(self, bin_edges: np.ndarray[tuple[int], np.dtype[np.floating]]) -> None:
        self.__bin_edges: np.ndarray[tuple[int], np.dtype[np.floating]] = bin_edges
        self.__bin_centres: np.ndarray[tuple[int], np.dtype[np.floating]] = (self.__bin_edges[:-1] + self.__bin_edges[1:]) / 2
        self.__bin_widths: np.ndarray[tuple[int], np.dtype[np.floating]] = self.__bin_edges[1:] - self.__bin_edges[:-1]

    def __len__(self) -> int:
        return len(self.__bin_centres)
    
    @property
    def edges(self) -> np.ndarray[tuple[int], np.dtype[np.floating]]:
        """
        Get the bin edges.

        Returns:
            np.ndarray[tuple[int], np.dtype[np.floating]] -> The bin edges.
        """
        return self.__bin_edges
    
    @property
    def centres(self) -> np.ndarray[tuple[int], np.dtype[np.floating]]:
        """
        Get the bin centres.

        Returns:
            np.ndarray[tuple[int], np.dtype[np.floating]] -> The bin centres.
        """
        return self.__bin_centres
    
    @property
    def widths(self) -> np.ndarray[tuple[int], np.dtype[np.floating]]:
        """
        Get the bin widths.

        Returns:
            np.ndarray[tuple[int], np.dtype[np.floating]] -> The bin widths.
        """
        return self.__bin_widths
    
    @classmethod
    def __from_cache_data__(cls, data: dict[str, Any]) -> "Bins":
        """
        Load the object from a cache target.

        Parameters:
            dict[str, Any] data:
                The data to load the object from.
        """
        return Bins(data["bin_edges"])

    def __get_cache_data__(self) -> dict[str, Any]:
        """
        Get the data to be cached.

        Returns:
            dict[str, Any] -> The data to be cached.
        """
        return { "bin_edges": self.__bin_edges }
    
    @staticmethod
    def make_linear_bins(min: float, max: float, number: int, centred_limits: bool = False) -> "Bins":
        """
        Create a set of linear bins.

        Parameters:
            min (float): The minimum value of the bins.
            max (float): The maximum value of the bins.
            number (int): The number of bins to create.
            centred_limits (bool): Whether the limits are for the centre of the outermost bins or the outside edges (default).

        Returns:
            Bins -> The linear bins.
        """
        bin_edges: np.ndarray[tuple[int], np.dtype[np.floating]] = np.linspace(min, max, number + 1) if not centred_limits else np.linspace(min - ((max - min) / (2 * number)), max + ((max - min) / (2 * number)), number + 1)
        return Bins(bin_edges)

    @staticmethod
    def make_logarithmic_bins(min: float, max: float, number: int, centred_limits: bool = False) -> "Bins":
        """
        Create a set of logarithmic bins.

        Parameters:
            min (float): The minimum value of the bins.
            max (float): The maximum value of the bins.
            number (int): The number of bins to create.
            centred_limits (bool): Whether the limits are for the centre of the outermost bins or the outside edges (default).

        Returns:
            Bins -> The logarithmic bins.
        """
        bin_edges: np.ndarray[tuple[int], np.dtype[np.floating]] = np.logspace(np.log10(min), np.log10(max), number + 1) if not centred_limits else np.logspace(np.log10(min - ((max - min) / (2 * number))), np.log10(max + ((max - min) / (2 * number))), number + 1)
        return Bins(bin_edges)



import pickle

from . import CheckpointBase

class PickleCheckpoint(CheckpointBase):
    """
    WARNING: This class uses pickle binary files. These are INSECURE and may be altered maliciously!
             Use with caution and never load pickle files that you didn't create yourself.
    """

    def _read_file(self, filepath):
        print("READING")
        data = None
        with open(filepath, "rb") as file:
            data = pickle.load(file)
        return data

    def _write_file(self, filepath, data, *args, **kwargs):
        with open(filepath, "wb") as file:
            pickle.dump(data, file)



from typing import Sequence, Literal

from matplotlib.font_manager import FontProperties
from matplotlib.typing import ColorType

from ..Tools._Struct import CacheableStruct
from ..Tools._autoproperty import AutoProperty

class CachedPlotFontInfo(CacheableStruct):
    size    = AutoProperty[float|Literal["xx-small","x-small","small","medium","large","x-large","xx-large"]](allow_uninitialised = True) # Matplotlib default is 10.0
    family  = AutoProperty[Sequence[str]](allow_uninitialised = True)
    style   = AutoProperty[str          ](allow_uninitialised = True)
    variant = AutoProperty[str          ](allow_uninitialised = True)
    weight  = AutoProperty[str|float    ](allow_uninitialised = True)
    stretch = AutoProperty[str          ](allow_uninitialised = True)
    colour  = AutoProperty[ColorType    ](allow_uninitialised = True)
    def __init__(self, **kwargs):
        super().__init__(
            cacheable_attributes = ["size", "family", "style", "variant", "weight", "stretch", "colour"],
            **kwargs
        )

    @staticmethod
    def from_font_properties(font_properties: FontProperties) -> "CachedPlotFontInfo":
        instance = CachedPlotFontInfo()
        instance.size    = font_properties.get_size()
        instance.family  = font_properties.get_family()
        instance.style   = font_properties.get_style()
        instance.variant = font_properties.get_variant()
        instance.weight  = font_properties.get_weight()
        instance.stretch = font_properties.get_stretch()
        return instance

    @property
    def fontdict(self) -> dict[str, object]:
        fontdict = {}
        if self.size is not None:
            fontdict["fontsize"] = self.size
        if self.family is not None:
            fontdict["fontfamily"] = self.family
        if self.style is not None:
            fontdict["fontstyle"] = self.style
        if self.variant is not None:
            fontdict["fontvariant"] = self.variant
        if self.weight is not None:
            fontdict["fontweight"] = self.weight
        if self.stretch is not None:
            fontdict["fontstretch"] = self.stretch
        return fontdict

    @property
    def fontproperties(self) -> FontProperties:
        kwargs = {}
        if self.size is not None:
            kwargs["size"] = self.size
        if self.family is not None:
            kwargs["family"] = self.family
        if self.style is not None:
            kwargs["style"] = self.style
        if self.variant is not None:
            kwargs["variant"] = self.variant
        if self.weight is not None:
            kwargs["weight"] = self.weight
        if self.stretch is not None:
            kwargs["stretch"] = self.stretch
        return FontProperties(**kwargs)
    
    def copy(self) -> "CachedPlotFontInfo":
        """
        Returns a shallow copy.
        """
        instance = CachedPlotFontInfo()
        if self.size is not None:
            instance.size = self.size
        if self.family is not None:
            instance.family = self.family
        if self.style is not None:
            instance.style = self.style
        if self.variant is not None:
            instance.variant = self.variant
        if self.weight is not None:
            instance.weight = self.weight
        if self.stretch is not None:
            instance.stretch = self.stretch
        return instance
    
    def with_default(self, default: "CachedPlotFontInfo|None") -> "CachedPlotFontInfo":
        """
        Return a new instance using the default values specified (where there are any).
        """
        if default is None:
            return self.copy()
        instance = default.copy()
        if self.size is not None:
            instance.size = self.size
        if self.family is not None:
            instance.family = self.family
        if self.style is not None:
            instance.style = self.style
        if self.variant is not None:
            instance.variant = self.variant
        if self.weight is not None:
            instance.weight = self.weight
        if self.stretch is not None:
            instance.stretch = self.stretch
        return instance



from ._CheckpointBase import CheckpointBase

class CheckpointState(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__initialised = False
        self.__checkpoint_index = -1
        self.__checkpoint_data = None

    @property
    def checkpoint_index(self):
        if not self.__initialised:
            raise RuntimeError("This object has not been initialised yet.")
        return self.__checkpoint_index

    @property
    def checkpoint_data(self):
        if not self.__initialised:
            raise RuntimeError("This object has not been initialised yet.")
        return self.__checkpoint_data

    @checkpoint_data.setter
    def checkpoint_data(self, value):
        if not self.__initialised:
            raise RuntimeError("This object has not been initialised yet.")
        self.__checkpoint_data = value

    def initialise(self, from_checkpoint: dict = None):
        if from_checkpoint is not None:
            self.__checkpoint_index = from_checkpoint["index"]
            self.__checkpoint_data = from_checkpoint["data"]
            for i in range(len(from_checkpoint["metadata_keys"])):
                self[from_checkpoint["metadata_keys"][i]] = from_checkpoint["metadata_values"][i]
        self.__initialised = True

    def serialise(self):
        serialised = {}
        serialised["index"] = self.__checkpoint_index
        serialised["data"] = self.__checkpoint_data
        serialised["metadata_keys"] = list(self.keys())
        serialised["metadata_values"] = list(self.values())
        return serialised

    def increment_index(self):
        self.__checkpoint_index += 1



from abc import ABC
from typing import TypeVar, Generic



T = TypeVar("T")

class _MPI_Config_Base(ABC, Generic[T]):
    __singleton: "_MPI_Config_Base|None" = None
    __update: bool = True

    @staticmethod
    def _is_singleton_avalible() -> bool:
        return _MPI_Config_Base.__singleton is not None

    @staticmethod
    def _get_singleton() -> "_MPI_Config_Base|None":
        return _MPI_Config_Base.__singleton
    
    def __init__(self, comm: T, size: int, rank: int, root: int) -> None:
        if _MPI_Config_Base.__update:
            _MPI_Config_Base.__update = False
            _MPI_Config_Base.__singleton = self
            self.__MPI_COMM: T = comm
            self.__MPI_COMM_SIZE: int = size
            self.__MPI_RANK: int = rank
            self.__MPI_ROOT_RANK: int = root

        else:
            raise RuntimeError("Only one instance of the MPI_Config object may exist. Change configuration using the update method.")

    @property
    def comm(self) -> T:
        return self.__MPI_COMM

    @property
    def comm_size(self) -> int:
        return self.__MPI_COMM_SIZE

    @property
    def rank(self) -> int:
        return self.__MPI_RANK

    @property
    def root(self) -> int:
        return self.__MPI_ROOT_RANK
    @root.setter
    def root(self, rank: int) -> None:
        if not isinstance(rank, int):
            raise TypeError(f"Type of argument provided for parameter \"rank\" is {type(rank)} not int.")
        if rank < 0 or rank >= self.comm_size:
            raise ValueError(f"Argument provided for parameter \"rank\" was {rank} which is outside the valid range 0 -> {self.comm_size - 1}")
        self.__MPI_ROOT_RANK = rank

    @property
    def is_root(self) -> bool:
        return self.rank == self.root
    def check_is_root(self, root: int|None = None) -> bool:
        """
        Equivilant of root_rank == current_rank but allows for the default case where root_rank is the configured root rank for the default communicator.
        """
        return self.rank == (self.root if root is None else root)

    def update(self, comm: T, root: int|None = None) -> None:
        _MPI_Config_Base.__update(comm, self.root if root is None else root)

    @classmethod
    def __update(cls, comm: T, root: int) -> None:
        _MPI_Config_Base.__update = True
        instance = cls(comm)
        instance.root = root

    def allow_default_comm(self, comm: T|None = None):
        return comm if comm is not None else self.comm

    def allow_default_root(self, root: int|None = None):
        return root if root is not None else self.root
